#!/usr/bin/env python
#
# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# btrfs module for System Storage Manager

import re
import os
import sys
import datetime
from ssmlib import misc
from ssmlib import problem

__all__ = ["BtrfsVolume", "BtrfsPool", "BtrfsDev"]

try:
    SSM_BTRFS_DEFAULT_POOL = os.environ['SSM_BTRFS_DEFAULT_POOL']
except KeyError:
    SSM_BTRFS_DEFAULT_POOL = "btrfs_pool"


def get_real_number(string):
    number = float(string[0:-2])
    unit = string[-2:-1]
    # The result will be in kilobytes
    units = ["K", "M", "G", "T", "P", "E", "Z", "Y"]
    for i, u in enumerate(units):
        if u == unit:
            number *= (2 ** (i * 10))
            break
    return number


def get_btrfs_version():
    try:
        output = misc.run(['btrfs', '--version'], can_fail=True)[1]
        output = output.strip().split("\n")[-1]
        version = re.search('(?<=v)\d+\.\d+', output).group(0)
    except (OSError, AttributeError):
        version = "0.0"
    return float(version)

BTRFS_VERSION = get_btrfs_version()


class Btrfs(object):

    def __init__(self, options, data=None):
        self.type = 'btrfs'
        self.data = data or {}
        self.options = options
        self.default_pool_name = SSM_BTRFS_DEFAULT_POOL
        self._vol = {}
        self._pool = {}
        self._dev = {}
        self._snap = {}
        self._subvolumes = {}
        self._binary = misc.check_binary('btrfs')
        self.problem = problem.ProblemSet(options)
        self.modified_list_version = True

        if not self._binary:
            return

        self.mounts = misc.get_mounts('/dev/')
        command = ['btrfs', 'filesystem', 'show']
        self.output = misc.run(command, stderr=False)[1]

        vol = {}
        pool = {}
        dev = {}
        partitions = {}
        fs_size = pool_size = fs_used = 0
        pool_name = ''
        for line in misc.get_partitions():
            partitions[line[3]] = line

        for line in self.output.strip().split("\n"):
            if not line:
                continue
            array = line.split()

            if array[0] == 'Label:':
                if len(vol) > 0:
                    self._store_data(vol, pool, fs_used, fs_size, pool_size,
                                     pool_name)
                    vol = {}
                    pool = {}
                    fs_size = pool_size = 0
                    pool_name = ''

                label = array[1].strip("'")
                uuid = array[3]
                pool['uuid'] = vol['uuid'] = uuid

                try:
                    fallback = False
                    vol['real_dev'] = misc.get_device_by_uuid(uuid)

                    if vol['real_dev'] in self.mounts:
                        pool['mount'] = self.mounts[vol['real_dev']]['mp']
                        vol['mount'] = self.mounts[vol['real_dev']]['mp']
                    else:
                        for dev_i in self.mounts:
                            found = re.findall(r'{0}:/.*'.format(vol['real_dev']), dev_i)
                            if found:
                                pool['mount'] = self.mounts[found[0]]['mp']
                                break
                except OSError:
                    # udev is "hard-to-work-with" sometimes so this is fallback
                    fallback = True
                    vol['real_dev'] = ""

                if label != 'none':
                    vol['label'] = label
                vol['ID'] = 0

            elif array[0] == 'Total':
                pool['dev_count'] = array[2]
                fs_used = get_real_number(array[6])

            elif array[0] == 'devid':
                dev['dev_name'] = array[7]

                if not pool_name:
                    pool_name = self._find_uniq_pool_name(label, array[7])
                dev['pool_name'] = pool_name

                # Fallback in case we could not find real_dev by uuid
                if fallback and 'mount' not in pool:
                    if dev['dev_name'] in self.mounts:
                        pool['mount'] = self.mounts[dev['dev_name']]['mp']
                        vol['real_dev'] = dev['dev_name']

                        if 'root' in self.mounts[dev['dev_name']]:
                            if self.mounts[dev['dev_name']]['root'] == '/':
                                vol['mount'] = self.mounts[dev['dev_name']]['mp']
                    else:
                        for dev_i in self.mounts:
                            found = re.findall(r'{0}:/.*'.format(dev['dev_name']), dev_i)
                            if found:
                                pool['mount'] = self.mounts[found[0]]['mp']
                                break

                dev_used = get_real_number(array[5])
                dev['dev_used'] = str(dev_used)
                fs_size += get_real_number(array[3])

                dev_size = \
                    int(partitions[dev['dev_name'].rpartition("/")[-1]][2])
                pool_size += dev_size
                dev['dev_free'] = dev_size - dev_used
                self._dev[dev['dev_name']] = dev
                dev = {}

        if len(vol) > 0:
            self._store_data(vol, pool, fs_used, fs_size, pool_size, pool_name)

    def run_btrfs(self, command):
        if not self._binary:
            self.problem.check(self.problem.TOOL_MISSING, 'btrfs')
        command.insert(0, "btrfs")
        return misc.run(command, stdout=True)

    def _list_subvolumes(self, mount, list_snapshots=False):
        command = ['btrfs', 'subvolume', 'list']
        if self.modified_list_version:
            command.append('-a')
        if list_snapshots:
            command.append('-s')
        ret, output = misc.run(command + [mount], stdout=False, can_fail=True)
        if ret:
            command = ['btrfs', 'subvolume', 'list']
            if list_snapshots:
                command.append('-s')
            output = misc.run(command + [mount], stdout=False)[1]
            self.modified_list_version = False
        return output

    # There is no way in btrfs to list subvolumes which are not snapshots
    # so we have to get the list of snapshots to filter it out from
    # regular subvolume list so we do not have it in the output twice.
    # Once in volume list and once in snapshot list.
    def _get_snap_name_list(self, mount):
        snap = []
        if BTRFS_VERSION < 0.20:
            return snap
        command = ['btrfs', 'subvolume', 'list', '-s', mount]
        output = misc.run(command, stdout=False)[1]

        for line in output.strip().split("\n"):
            if not line:
                continue
            path = re.search('(?<=path ).*$', line).group(0)
            snap.append(path)
        return snap

    def _fill_subvolumes(self, list_snapshots=False):
        if not self._binary:
            return
        if self._subvolumes:
            return
        for name, vol in self._vol.iteritems():
            pool_name = vol['pool_name']
            real_dev = vol['real_dev']
            pool = self._pool[pool_name]

            if 'mount' in self._pool[pool_name]:
                mount = pool['mount']
            else:
                # If btrfs is not mounted we will not process subvolumes
                continue

            snapshots = []
            if not list_snapshots:
                snapshots = self._get_snap_name_list(mount)

            output = self._list_subvolumes(mount, list_snapshots)
            for volume in self._parse_subvolumes(output):
                new = vol.copy()
                new.update(volume)
                new['dev_name'] = "{0}:{1}".format(name, new['path'])
                item = "{0}:/{1}".format(real_dev, new['path'])
                # If the subvolume is mounted we should find it here
                if item in self.mounts:
                    new['mount'] = self.mounts[item]['mp']
                    # Subvolume is mounted directly
                    new['direct_mount'] = True
                else:
                    # If subvolume is not mounted try to find whether parent
                    # subvolume is mounted
                    found = re.findall(r'^(.*)/([^/]*)$', new['path'])
                    if found:
                        parent_path, path = found[0]
                        # try previously loaded subvolumes
                        for prev_sv in self._subvolumes:
                            # if subvolumes are mounted, use that mp
                            if self._subvolumes[prev_sv]['path'] == parent_path:
                                # if parent subvolume is not mounted this
                                # subvolume is not mounted as well
                                if self._subvolumes[prev_sv]['mount'] == '':
                                    new['mount'] = ''
                                else:
                                    new['mount'] = "{0}/{1}".format(
                                        self._subvolumes[prev_sv]['mount'],
                                        path)
                                break
                    # if parent volume is not mounted, use root subvolume
                    # if mounted
                    else:
                        if 'mount' in vol:
                            new['mount'] = "{0}/{1}".format(vol['mount'],
                                                            new['path'])

                new['hide'] = False
                # Store snapshot info
                if 'mount' in new and \
                    re.match("snap-\d{4}-\d{2}-\d{2}-T\d{6}",
                             os.path.basename(new['mount'])):
                    new['snap_name'] = "{0}:{1}".format(name,
                            os.path.basename(new['path']))
                    new['snap_path'] = new['mount']
                if volume['path'] in snapshots:
                    new['hide'] = True

                self._subvolumes[new['dev_name']] = new

    def _parse_subvolumes(self, output):
        volume = {}
        for line in output.strip().split("\n"):
            if not line:
                continue
            # For the version with screwed 'subvolume list' command
            line = re.sub("<FS_TREE>/*", "", line)
            volume['ID'] = re.search('(?<=ID )\d+', line).group(0)
            volume['top_level'] = re.search('(?<=top level )\d+', line).group(0)
            volume['path'] = re.search('(?<=path ).*$', line).group(0)
            volume['subvolume'] = True
            yield volume

    def _find_uniq_pool_name(self, label, dev):
        if len(label) < 3 or label == "none":
            label = "btrfs_{0}".format(os.path.basename(dev))
        if label not in self._pool:
            return label
        return os.path.basename(dev)

    def _store_data(self, vol, pool, fs_used, fs_size, pool_size, pool_name):
        vol['fs_type'] = 'btrfs'
        vol['fs_used'] = str(fs_used)
        vol['fs_free'] = str(fs_size - fs_used)
        vol['fs_size'] = vol['vol_size'] = pool['pool_used'] = \
            str(fs_size)
        pool['pool_free'] = str(pool_size - fs_used)
        pool['pool_size'] = pool_size
        pool['pool_name'] = vol['pool_name'] = vol['dev_name'] = pool_name
        pool['type'] = 'btrfs'
        vol['type'] = 'btrfs'

        self._pool[pool['pool_name']] = pool
        self._vol[vol['dev_name']] = vol

    def __iter__(self):
        for item in sorted(self.data.iterkeys()):
            yield item

    def __getitem__(self, key):
        if key in self.data.iterkeys():
            return self.data[key]

    def _remove_filesystem(self, name):
        if 'mount' in self._vol[name]:
            if self.problem.check(self.problem.FS_MOUNTED,
                                  [name, self._vol[name]['mount']]):
                misc.do_umount(self._vol[name]['mount'])
        for dev in self._dev.itervalues():
            if dev['pool_name'] != name:
                continue
            misc.wipefs(dev['dev_name'], 'btrfs')


class BtrfsVolume(Btrfs):

    def __init__(self, *args, **kwargs):
        super(BtrfsVolume, self).__init__(*args, **kwargs)
        self._fill_subvolumes()
        if self.data:
            self.data.update(self._vol)
            self.data.update(self._subvolumes)
        else:
            self.data = self._vol
            self.data.update(self._subvolumes)

    def mount(self, vol, mpoint, options=None):
        vol = self.data[vol]
        if options:
            options += ","
        else:
            options = ""
        options += "subvolid={0}".format(vol['ID'])
        misc.do_mount(vol['real_dev'], mpoint, options)

    def remove(self, vol):
        volume = self._vol[vol]
        if 'subvolume' in volume:
            # If subvolume is mounted directly we can not remove it. So ask
            # user whether he wants to umount it. The we'll have to mount the
            # root subvolume and remove this subvolume.
            if 'direct_mount' in volume and volume['direct_mount']:
                if self.problem.check(self.problem.FS_MOUNTED,
                                      [vol, volume['mount']]):
                    misc.do_umount(volume['mount'])
                    del volume['mount']
                    del volume['direct_mount']
            if 'mount' not in volume:
                mount = misc.temp_mount("UUID={0}".format(volume['uuid']))
                path = "{0}/{1}".format(mount, volume['path'])
            else:
                path = volume['mount']
            self.run_btrfs(['subvolume', 'delete', path])
        else:
            self._remove_filesystem(vol)

    def resize(self, vol, size, resize_fs=True):
        vol = self.data[vol]
        if 'mount' not in vol:
            tmp = misc.temp_mount("UUID={0}".format(vol['uuid']))
            vol['mount'] = tmp
        command = ['filesystem', 'resize', str(int(size)) + "K", vol['mount']]
        self.run_btrfs(command)

    def snapshot(self, vol, destination, name, size, user_set_size):
        vol = self.data[vol]
        if 'mount' not in vol:
            tmp = misc.temp_mount("UUID={0}".format(vol['uuid']))
            vol['mount'] = tmp

        if not destination and not name:
            now = datetime.datetime.now()
            destination = vol['mount'] + now.strftime("/snap-%Y-%m-%d-T%H%M%S")
        if name:
            destination = vol['mount'] + "/" + name

        if user_set_size:
            self.problem.warn("Btrfs doesn't allow setting a size of " +
                              "subvolumes")

        command = ['subvolume', 'snapshot', vol['mount'], destination]
        self.run_btrfs(command)


class BtrfsDev(Btrfs):

    def __init__(self, *args, **kwargs):
        super(BtrfsDev, self).__init__(*args, **kwargs)
        if self.data:
            self.data.update(self._dev)
        else:
            self.data = self._dev

    def remove(self, devices):
        raise Exception("Not sure what you want to" +
                        "achieve by removing {0}".format(devices))


class BtrfsPool(Btrfs):

    def __init__(self, *args, **kwargs):
        super(BtrfsPool, self).__init__(*args, **kwargs)
        if self.data:
            self.data.update(self._pool)
        else:
            self.data = self._pool

    def _create_filesystem(self, pool, name, devs, size=None, raid=None):
        if not devs:
            raise Exception("To create btrfs volume, some devices must be " +
                            "provided")
        self._binary = misc.check_binary('mkfs.btrfs')
        if not self._binary:
            self.problem.check(self.problem.TOOL_MISSING, 'mkfs.btrfs')
        command = ['mkfs.btrfs', '-L', name]

        if raid:
            if raid['level'] == '0':
                command.extend(['-m', 'raid0', '-d', 'raid0'])
            elif raid['level'] == '1':
                command.extend(['-m', 'raid1', '-d', 'raid1'])
            elif raid['level'] == '10':
                command.extend(['-m', 'raid10', '-d', 'raid10'])
            else:
                raise Exception("Btrfs backed currently does not support " +
                                "RAID level {0}".format(raid['level']))

        if size:
            command.extend(['-b', "{0}".format(int(float(size) * 1024))])
        # This might seem weird, but btrfs is mostly broken when it comes to
        # checking existing signatures because it will for example check for
        # backup superblocks as well, which is wrong. Also we have check for
        # existing file system signatures in the ssm itself. Other things
        # than file system should be covered by the backend and we should
        # have tried to remove the device from the respective pool already.
        # So at this point there should not be any useful signatures to
        # speak of. However as I mentioned btrfs is broken, so force it.
        command.extend(['-f'])
        command.extend(devs)
        misc.run(command, stdout=True)
        misc.send_udev_event(devs[0], "change")
        return name

    def _check_new_path(self, path, name):
        msg = 0
        parent = os.path.split(path)[0]
        if os.path.exists(path):
            msg = "Directory \'{0}\' already exist. ".format(path) + \
                  "Subvolume \'{0}\' can not be ".format(name) + \
                  "created"
        elif not os.path.exists(parent):
            msg = "Parent directory \'{0}\' ".format(parent) + \
                  "does not exist. Subvolume " + \
                  "\'{0}\' can not be created".format(name)
        if msg:
            self.problem.error(msg)

    def reduce(self, pool, device):
        pool = self.data[pool]
        if 'mount' not in pool:
            tmp = misc.temp_mount("UUID={0}".format(pool['uuid']))
            pool['mount'] = tmp
        command = ['device', 'delete', device, pool['mount']]
        self.run_btrfs(command)
        misc.send_udev_event(device, "change")

    def new(self, pool, devices):
        if type(devices) is not list:
            devices = [devices]
        self.create(pool, devs=devices)

    def extend(self, pool, devices):
        pool = self.data[pool]
        if 'mount' not in pool:
            tmp = misc.temp_mount("UUID={0}".format(pool['uuid']))
            pool['mount'] = tmp
        if type(devices) is not list:
            devices = [devices]
        command = ['device', 'add']
        command.extend(devices)
        command.append(pool['mount'])
        self.run_btrfs(command)

    def remove(self, pool):
        # Volume and pool name should be the same, since it actually is the
        # same file system
        self._remove_filesystem(pool)

    def create(self, pool, size=None, name=None, devs=None,
               raid=None):
        if pool in self._pool:
            vol = None
            if size or raid:
                self.problem.warn("Only name, volume name and pool name " +
                                  "can be specified when creating btrfs " +
                                  "subvolume, the rest will be ignored")
            tmp = misc.temp_mount("UUID={0}".format(self._pool[pool]['uuid']))
            self._pool[pool]['mount'] = tmp

            if not name:
                now = datetime.datetime.now()
                name = now.strftime("%Y-%m-%d-T%H%M%S")
                vol = "{0}/{1}".format(self._pool[pool]['mount'], name)
            elif os.path.isabs(name):
                vol = name
            else:
                vol = "{0}/{1}".format(self._pool[pool]['mount'], name)

            self._check_new_path(vol, name)
            self.run_btrfs(['subvolume', 'create', vol])
            vol = "{0}:{1}".format(pool, name)
        else:
            if len(devs) == 0:
                self.problem.check(self.problem.NO_DEVICES, pool)
            if name:
                self.problem.warn("Creating new pool. Argument (--name " +
                                  "{0}) will be ignored!".format(name))
            vol = self._create_filesystem(pool, pool, devs, size, raid)
        return vol


class BtrfsSnap(Btrfs):

    def __init__(self, *args, **kwargs):
        super(BtrfsSnap, self).__init__(*args, **kwargs)

        self._fill_subvolumes(list_snapshots=True)
        for name, vol in self._subvolumes.iteritems():
            if BTRFS_VERSION < 0.20:
                if 'snap_name' in vol:
                    self._snap[vol['snap_name']] = vol.copy()
                    self._snap[vol['snap_name']]['hide'] = False
            else:
                self._snap[vol['dev_name']] = vol.copy()

        if self.data:
            self.data.update(self._snap)
        else:
            self.data = self._snap



#!/usr/bin/env python
#
# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# System Storage Manager - ssm

import re
import os
import sys
import stat
import argparse
from ssmlib import misc
from ssmlib import problem

# Import backends
from ssmlib.backends import lvm, crypt, btrfs, md

EXTN = ['ext2', 'ext3', 'ext4']
SUPPORTED_FS = ['xfs', 'btrfs'] + EXTN
SUPPORTED_BACKENDS = ['lvm', 'btrfs']
SUPPORTED_RAID = ['0', '1', '10']
os.environ['LC_NUMERIC'] = "C"

# If you change this please change doc/conf.py as well
VERSION = '0.4'

# Should the script be run in interactive or non interactive mode ?
try:
    SSM_NONINTERACTIVE = os.environ['SSM_NONINTERACTIVE']
    if SSM_NONINTERACTIVE.upper() in ['YES', 'TRUE', '1']:
        SSM_NONINTERACTIVE = True
    elif SSM_NONINTERACTIVE.upper() in ['NO', 'FALSE', '0']:
        SSM_NONINTERACTIVE = False
except KeyError:
    SSM_NONINTERACTIVE = not os.isatty(sys.stdout.fileno())


class Options(object):
    """
    Structure that contains option setting allowing it to be
    passed between parts of ssm.
    """

    def __init__(self):
        self.interactive = not SSM_NONINTERACTIVE
        self.verbose = False
        self.debug = False
        self.force = False
        self.yes = False
        self.config = None


# Initialize problem set
PR = problem.ProblemSet(Options())

# Name of the default pool
try:
    DEFAULT_DEVICE_POOL = os.environ['DEFAULT_DEVICE_POOL']
except KeyError:
    DEFAULT_DEVICE_POOL = "device_pool"

# Default back-end
try:
    SSM_DEFAULT_BACKEND = os.environ['SSM_DEFAULT_BACKEND']
    if SSM_DEFAULT_BACKEND not in SUPPORTED_BACKENDS:
        if PR.check(PR.BAD_ENV_VARIABLE,
                    ['SSM_DEFAULT_BACKEND', SSM_DEFAULT_BACKEND]):
            SSM_DEFAULT_BACKEND = 'lvm'
except KeyError:
    SSM_DEFAULT_BACKEND = 'lvm'


# If this environment variable is set, ssm will only consider such devices,
# pools and volumes which names start with this prefix. This is especially
# useful for testing.
try:
    SSM_PREFIX_FILTER = os.environ['SSM_PREFIX_FILTER']
    PR.warn("SSM_PREFIX_FILTER is set to \'{0}\'".format(SSM_PREFIX_FILTER))
except KeyError:
    SSM_PREFIX_FILTER = None


class Struct(object):
    def __init__(self):
        pass


class StoreAll(argparse._StoreAction):
    """
    Argparse class used to store all valid values. Valid values should not be
    empty or None
    """

    def __call__(self, parser, namespace, values, option_string=None):
        for val in values[:]:
            if not val:
                values.remove(val)
        setattr(namespace, self.dest, values)


class SetBackend(argparse._StoreAction):
    """
    Action for the backend parameter, where we want to store provided
    in SSM_DEFAULT_BACKEND.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        # Set default backend to the provided value. All check should be
        # already done by argparse.
        global SSM_DEFAULT_BACKEND
        SSM_DEFAULT_BACKEND = values[0]
        setattr(namespace, self.dest, values)


class FsInfo(object):
    """
    Parse and store information about the file system. Methods specific for
    each file system should be part of this class
    """

    def __init__(self, dev, options):
        self.data = {}
        self.options = options
        fstype = misc.get_fs_type(dev)
        if fstype not in [None, 'btrfs']:
            self.data['fs_type'] = fstype
        else:
            return

        self.fs_info = {}
        if fstype in EXTN:
            self.extN_get_info(dev)
        elif fstype == "xfs":
            self.xfs_get_info(dev)
        self.fstype = fstype
        self.device = dev
        self.mounted = False

    def _get_fs_func(self, func, *args, **kwargs):
        fstype = self.fstype
        if re.match("ext[2|3|4]", self.fstype):
            fstype = "extN"
        func = getattr(self, "{0}_{1}".format(fstype, func))
        return func(*args, **kwargs)

    def fsck(self):
        return self._get_fs_func("fsck")

    def resize(self, *args, **kwargs):
        return self._get_fs_func("resize", *args, **kwargs)

    def get_info(self, *args, **kwargs):
        return self._get_fs_func("get_info", *args, **kwargs)

    def extN_get_info(self, dev):
        command = ["tune2fs", "-l", dev]
        if not misc.check_binary(command[0]):
            return
        output = misc.run(command)[1]

        for line in output.split("\n")[1:]:
            array = line.split(":")
            if len(array) == 2:
                self.fs_info[array[0]] = array[1].lstrip()

        bsize = int(self.fs_info['Block size'])
        bcount = int(self.fs_info['Block count'])
        rbcount = int(self.fs_info['Reserved block count'])
        fbcount = int(self.fs_info['Free blocks'])
        self.data['fs_size'] = bcount * bsize / 1024
        self.data['fs_free'] = (fbcount - rbcount) * bsize / 1024
        self.data['fs_used'] = (bcount - fbcount) * bsize / 1024

    def extN_fsck(self):
        command = ['fsck.{0}'.format(self.fstype), '-f', '-n']
        if not misc.check_binary(command[0]):
            PR.warn("\'{0}\' tool does not exist. ".format(command[0]) +
                    "File system will not be checked")
            return 1
        if self.options.force:
            command.append('-f')
        if self.options.verbose:
            command.append('-v')
        command.append(self.device)
        return misc.run(command, stdout=True, can_fail=True)[0]

    def extN_resize(self, new_size=None):
        command = ['resize2fs', self.device]
        if not misc.check_binary(command[0]):
            PR.warn("\'{0}\' tool does not exist. ".format(command[0]) +
                    "File system will not be resized")
            return 1
        if self.options.force:
            command.insert(1, "-f")
        if self.options.verbose:
            command.insert(1, "-p")
        if new_size:
            command.append(new_size)
        # Ext3/4 can resize offline in both directions, but It can not shrink
        # the file system while online. In addition ext2 can only resize
        # offline.
        if self.mounted and (self.fstype == "ext2" or
           new_size < self.data['fs_size']):
            raise Exception(
                "{0} is mounted on {1}".format(self.device, self.mounted) +
                " In this case, mounted file system can not be resized.")
        ret = self.fsck()
        if ret:
            raise Exception("File system on {0} is not ".format(self.device) +
                            "clean, I will not attempt to resize it. Please," +
                            "fix the problem first.")
        misc.run(command, stdout=True)

    def xfs_get_info(self, dev):
        command = ["xfs_db", "-r", "-c", "sb", "-c", "print", dev]
        if not misc.check_binary(command[0]):
            return
        output = misc.run(command)[1]

        for line in output.split("\n")[1:]:
            array = line.split("=")
            if len(array) == 2:
                self.fs_info[array[0].rstrip()] = array[1].lstrip()

        bsize = int(self.fs_info['blocksize'])
        bcount = int(self.fs_info['dblocks'])
        lbcount = int(self.fs_info['logblocks'])
        bcount -= lbcount
        agcount = int(self.fs_info['agcount'])
        fbcount = int(self.fs_info['fdblocks'])
        fbcount -= 4 + (4 + agcount)
        self.data['fs_size'] = bcount * bsize / 1024
        self.data['fs_free'] = fbcount * bsize / 1024
        self.data['fs_used'] = (bcount - fbcount) * bsize / 1024

    def xfs_fsck(self):
        command = ['xfs_check']
        if not misc.check_binary(command[0]):
            PR.warn("\'{0}\' tool does not exist. ".format(command[0]) +
                    "File system will not be checked")
            return 1
        if self.options.verbose:
            command.append('-v')
        command.append(self.device)
        return misc.run(command, stdout=True, can_fail=True)[0]

    def xfs_resize(self, new_size=None):
        command = ['xfs_growfs', self.device]
        if not misc.check_binary(command[0]):
            PR.warn("\'{0}\' tool does not exist. ".format(command[0]) +
                    "File system will not be resized")
            return 1
        if new_size:
            command.insert(1, ['-D', new_size + 'K'])
        if not self.mounted:
            raise Exception("Xfs file system on {0}".format(self.device) +
                            " has to be mounted to perform an resize.")
        elif new_size and new_size < self.data['fs_size']:
            raise Exception("Xfs file system can not shrink.")
        else:
            misc.run(command, stdout=True)


class DeviceInfo(object):
    """
    Parse and store information about the devices present in the system. The
    main source of information are /proc/partitions, /proc/mounts and
    /proc/swaps. self.data should be appended to since there might be other
    data present which will add more information about devices, usually
    provided from backends.

    Important thing is that we hide all dm devices here, since they might
    really be a volumes. We let backend decide whether the device should be
    listed as device or not simply by setting 'hide' to True/False.
    """

    def __init__(self, options, data=None):
        self.type = 'device'
        self.data = data or {}
        self.attrs = ['major', 'minor', 'dev_size', 'dev_name']
        self.options = options

        hide_dmnumbers = []
        for name in ['device-mapper', 'sr', 'md']:
            hide_dmnumbers.append(misc.get_dmnumber(name))

        mounts = misc.get_mounts('/dev/')
        swaps = misc.get_swaps()

        for items in misc.get_partitions():
            devices = dict(zip(self.attrs, items))
            devices['vol_size'] = devices['dev_size']
            devices['dev_name'] = "/dev/" + devices['dev_name']
            if devices['major'] in hide_dmnumbers:
                devices['hide'] = True
            if devices['dev_name'] in self.data:
                if 'hide' in self.data[devices['dev_name']] and \
                   not self.data[devices['dev_name']]['hide']:
                    devices['hide'] = False
                self.data[devices['dev_name']].update(devices)
            else:
                self.data[devices['dev_name']] = devices
            if devices['dev_name'] in mounts:
                devices['mount'] = mounts[devices['dev_name']]['mp']

        for item in swaps:
            if item[0] in self.data:
                self.data[item[0]]['mount'] = "SWAP"

        for i, dev in enumerate(self.data.itervalues()):
            if 'minor' in dev and dev['minor'] != '0':
                continue
            part = 0
            for a, d in enumerate(self.data.values()):
                if a == i:
                    continue
                try:
                    if dev['major'] != d['major']:
                        continue
                except KeyError:
                    continue
                if re.search(dev['dev_name'], d['dev_name']):
                    d['partition'] = True
                    d['type'] = 'part'
                    part += 1
            dev['partitioned'] = part
            if part > 0:
                dev['mount'] = "PARTITIONED"
                dev['type'] = 'disk'

    def set_globals(self, options):
        self.options = options

    def __iter__(self):
        for item in sorted(self.data.iterkeys()):
            yield item

    def __getitem__(self, name):
        device = misc.get_real_device(name)
        if device in self.data.iterkeys():
            return self.data[device]
        return None


class Item(object):
    """
    Meta object which provides encapsulation for all devices, pools and
    volumes, so we can work with them as with the usual objects without the
    need to call Dev, Pool or Vol methods directly.
    """

    def __init__(self, obj, name):
        self.obj = obj
        self.name = name
        self.type = obj.type

    @property
    def data(self):
        return self.obj[self.name]

    def __getattr__(self, func_name):
        func = getattr(self.obj, func_name)

        def _new_func(*args, **kwargs):
            if args and kwargs:
                return func(self.name, *args, **kwargs)
            elif kwargs:
                return func(self.name, **kwargs)
            elif args:
                return func(self.name, *args)
            else:
                return func(self.name)

        return _new_func

    def __getitem__(self, key):
        if key not in self.data and \
           re.match(r"fs_.*", key):
            self._fill_fs_info()
        try:
            ret = self.data[key]
        except KeyError:
            ret = ""
        return ret

    def __contains__(self, item):
        if self[item]:
            return True
        else:
            return False

    def _fill_fs_info(self):
        if 'dm_name' in self.data:
            name = self.data['dm_name']
        elif 'real_dev' in self.data:
            name = self.data['real_dev']
        else:
            name = self.data['dev_name']
        fs = FsInfo(name, self.obj.options)
        try:
            fs.mounted = self.data['mount']
        except KeyError:
            fs.mounted = ""
        self.data.update(fs.data)
        self.data['fs_info'] = fs

    def exists(self):
        if self.name in self.obj:
            return True
        else:
            return False


class Storage(object):
    """
    Template class to use for storing information about Pools, Volumes and
    Devices from different backends. This simplify things a lot since we do not
    have to manually walk through all the backends, but this class will do this
    for us.
    """

    def __init__(self, options):
        self._data = {}
        self.header = None
        self.attrs = None
        self.types = None
        self.set_globals(options)

    def __iter__(self):
        for source in self._data.itervalues():
            for item in source:
                yield Item(source, item)

    def __contains__(self, item):
        if self[item]:
            return True
        else:
            return False

    def __getitem__(self, name):
        for source in self._data.itervalues():
            item = source[name]
            if item:
                return Item(source, name)
        return None

    def reinitialize(self):
        self.__init__(self.options)

    def _apply_prefix_filter(self):
        """
        If SSM_PREFIX FILTER is set, remove all items which basenames does not
        start with SSM_PREFIX_FILTER prefix. This is useful especially for
        testing so that ssm see only relevant devices and does not screw real
        system storage configuration.
        """
        if not SSM_PREFIX_FILTER:
            return
        reg = re.compile("^{0}".format(SSM_PREFIX_FILTER))
        for source in self._data.itervalues():
            for item in source:
                if reg.search(os.path.basename(item)):
                    continue
                if 'pool_name' in source.data[item] and \
                   reg.search(source.data[item]['pool_name']):
                    continue
                if 'dm_name' in source.data[item] and \
                   reg.search(os.path.basename(source.data[item]['dm_name'])):
                    continue
                del source.data[item]

    def get_backend(self, name):
        return self._data[name]

    def set_globals(self, options):
        self.options = options
        if self._data is None:
            return
        for source in self._data.itervalues():
            source.options = options

    def filesystems(self):
        for item in self:
            if 'fs_type' in item:
                yield item

    def ptable(self, cond=None, more_data=None, cond_func=None):
        """
        Print information table about the source (devices, pools, volumes)
        using the predefined variables (below). cond, or cond_func can be
        provided to decide which items not to print out.

        self.header - list of headers for the table
        self.attrs - list of attribute keys to print out
        self.types - types of the attributes to print out (str, or float/int)
        """
        lines = []
        fmt = ""

        if cond == "fs_only":
            iterator = self.filesystems()
        else:
            iterator = self

        # Keep track of used columns. Then we only print out columns with
        # values.
        columns = [False] * len(self.attrs)

        len_matrix = []
        index = 0
        # Gather all lines which are going to be printed into the list
        # and create matrix of attribute lengths.
        # Iterate through all items in each data source first.
        for data in misc.chain(iterator, more_data or []):
            if (cond_func and not cond_func(data)) or 'hide' in data:
                continue
            len_matrix.append([len(self.header[i])
                              for i in range(len(self.header))])
            line = ()
            # Iterate through all attributes in each item
            for i, attr in enumerate(self.attrs):
                if self.types[i] in (float, int):
                    item = misc.humanize_size(data[attr])
                elif attr + "_print" in data:
                    item = data[attr + "_print"]
                else:
                    item = data[attr]
                len_matrix[index][i] = len(item)
                line += item,
                if len(item) > 0:
                    columns[i] = True
            lines.append(line)
            index += 1

        if len(lines) == 0:
            return

        header = [item for item in misc.compress(self.header, columns)]
        alignment = list([(len(self.header[i]))
                         for i in range(len(self.header))])
        term_width = misc.terminal_size()[0]

        # Update matrix of attribute lengths and construct the final list
        # of alignment for each column in the table.
        for index in range(len(len_matrix)):
            line = None
            # Find maximum length for each column
            for a, array in enumerate(len_matrix):
                for i, item in enumerate(array):
                    if not columns[i]:
                        alignment[i] = 0
                        continue
                    if item > alignment[i]:
                        alignment[i] = item
                        line = a

            # Check the overall line length and if it is longer then the
            # actual terminal width we can wrap the line right after the
            # first attribute. Simply set the alignment to the smaller
            # possible and let recalculate the list of column alignments.
            # Note that when even with the line wrap we would still exceed
            # the terminal width, then there is nothing we can do about it
            # so do not bother with line wrapping at all since it would
            # only screw the formatting even more.
            length = sum(alignment) + 2 * len(header) - 2
            if length > term_width and \
               (length - term_width) < (alignment[0] - len(header[0])) and \
               line is not None:
                    alignment[0] = len(header[0])
                    len_matrix[line][0] = len(header[0])
            else:
                break

        # Get the actual line width
        width = sum(misc.compress(alignment, columns)) + 2 * len(header) - 2

        pos = 0
        # Use column alignments list to construct formatting string for each
        # line in the table. Note that some lines might be wrapped later on.
        for i, t in enumerate(self.types):
            if not columns[i]:
                continue
            if t in (float, int):
                fmt += "{{{0}:>{1}}}  ".format(pos, alignment[i])
            else:
                # Do not append additional spaces if this is the last item
                if i == len(header) - 1:
                    fmt += "{{{0}:{1}}}".format(pos, alignment[i])
                else:
                    fmt += "{{{0}:{1}}}  ".format(pos, alignment[i])
            pos += 1

        print "-" * width
        print fmt.format(*tuple(header))
        print "-" * width
        # Now print each line of the table. When the first attribute of the
        # line is longer than it should be we know that we have to wrap the
        # line.
        for i, line in enumerate(lines):
            line = misc.compress(line, columns)
            tmp1 = line.next()
            if len(tmp1) > alignment[0]:
                print tmp1
                print fmt.format('', *line)
            else:
                print fmt.format(tmp1, *line)
        print "-" * width


class Pool(Storage):
    """
    Store Pools from all the backends. When new backend is added into the ssm
    it should be registered withing this class with appropriate name.
    """

    def __init__(self, *args, **kwargs):
        super(Pool, self).__init__(*args, **kwargs)

        try:
            self._data['lvm'] = lvm.VgsInfo(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about LVM pools")
        try:
            self._data['btrfs'] = btrfs.BtrfsPool(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about btrfs pools")

        backend = self.get_backend(SSM_DEFAULT_BACKEND)
        self.default = Item(backend, backend.default_pool_name)
        self.header = ['Pool', 'Type', 'Devices', 'Free', 'Used', 'Total']
        self.attrs = ['pool_name', 'type', 'dev_count', 'pool_free',
                      'pool_used', 'pool_size']
        self.types = [str, str, str, float, float, float]
        self._apply_prefix_filter()


class Devices(Storage):
    """
    Store Devices from all the backends. When new backend is added into the ssm
    it should be registered withing this class with appropriate name.

    If the backend only have new information about the device which is already
    discovered by the DeviceInfo() class then it should just add the
    information into the existing devices by passing the data. But if the
    backed discovers new devices, it should add them as a new entry.
    """

    def __init__(self, *args, **kwargs):
        super(Devices, self).__init__(*args, **kwargs)

        try:
            my_lvm = lvm.PvsInfo(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about LVM physical volumes")
        try:
            my_btrfs = btrfs.BtrfsDev(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about btrfs devices")
            my_btrfs = Struct()
            my_btrfs.data = {}
        try:
            my_md = md.MdRaidDevice(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about MD devices")
            my_md = Struct()
            my_md.data = {}

        self._data['dev'] = DeviceInfo(data=dict(my_lvm.data.items() +
                                                 my_btrfs.data.items() +
                                                 my_md.data.items()),
                                       options=self.options)
        self.header = ['Device', 'Free', 'Used',
                       'Total', 'Pool', 'Mount point']
        self.attrs = ['dev_name', 'dev_free', 'dev_used', 'dev_size',
                      'pool_name', 'mount']
        self.types = [str, float, float, float, str, str]
        self._apply_prefix_filter()


class Volumes(Storage):
    """
    Store Volumes from all the backends. When new backend is added into the ssm
    it should be registered withing this class with appropriate name.
    """

    def __init__(self, *args, **kwargs):
        super(Volumes, self).__init__(*args, **kwargs)

        try:
            self._data['lvm'] = lvm.LvsInfo(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about LVM volumes")
        try:
            self._data['crypt'] = crypt.DmCryptVolume(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about crypt volumes")
        try:
            self._data['btrfs'] = btrfs.BtrfsVolume(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about btrfs volumes")
        try:
            self._data['md'] = md.MdRaidVolume(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about md raid volumes")

        self.header = ['Volume', 'Pool', 'Volume size', 'FS', 'FS size',
                       'Free', 'Type', 'Mount point']
        self.attrs = ['dev_name', 'pool_name', 'vol_size', 'fs_type',
                      'fs_size', 'fs_free', 'type', 'mount']
        self.types = [str, str, float, str, float, float, str, str]
        self._apply_prefix_filter()


class Snapshots(Storage):
    """
    Store Snapshots from all the backends that supports snapshotting. When
    the snapshotting support is added into the backed it should be registered
    within this class with appropriate name.
    """

    def __init__(self, *args, **kwargs):
        super(Snapshots, self).__init__(*args, **kwargs)

        try:
            self._data['lvm'] = lvm.SnapInfo(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about LVM snapshots")
        try:
            self._data['btrfs'] = btrfs.BtrfsSnap(options=self.options)
        except RuntimeError, err:
            PR.warn(err)
            PR.warn("Can not get information about btrfs snapshots")

        self.header = ['Snapshot', 'Origin', 'Pool', 'Volume size', 'Size',
                       'Type', 'Mount point']
        self.attrs = ['dev_name', 'origin', 'pool_name', 'vol_size',
                      'snap_size', 'type', 'mount']
        self.types = [str, str, str, float, float, str, str]
        self._apply_prefix_filter()


class StorageHandle(object):
    """
    The main class where all the magic is done. All the commands provided by
    ssm have its appropriate functions here which are then called by argparse.
    """

    def __init__(self, options=Options()):
        self._mpoint = None
        self._dev = None
        self._pool = None
        self._volumes = None
        self._snapshots = None
        self.set_globals(options)
        self.options = options

    def set_globals(self, options):
        if self._dev:
            self.dev.set_globals(options)
        if self._volumes:
            self.vol.set_globals(options)
        if self._pool:
            self.pool.set_globals(options)
        if self._snapshots:
            self.snap.set_globals(options)
        self.options = options

    @property
    def dev(self):
        if self._dev:
            return self._dev
        self._dev = Devices(options=self.options)
        return self._dev

    def reinit_dev(self):
        if self._dev:
            self._dev.reinitialize()

    @property
    def pool(self):
        if self._pool:
            return self._pool
        self._pool = Pool(options=self.options)
        return self._pool

    def reinit_pool(self):
        if self._pool:
            self._pool.reinitialize()

    @property
    def vol(self):
        if self._volumes:
            return self._volumes
        self._volumes = Volumes(options=self.options)
        return self._volumes

    def reinit_vol(self):
        if self._volumes:
            self._volumes.reinitialize()

    @property
    def snap(self):
        if self._snapshots:
            return self._snapshots
        self._snapshots = Snapshots(options=self.options)
        return self._snapshots

    def reinit_snap(self):
        if self._snapshots:
            self._snapshots.reinitialize()

    def _create_fs(self, fstype, volume):
        """
        Create a file system 'fstype' on the 'volume'.
        """
        command = ["mkfs.{0}".format(fstype), volume]
        if not misc.check_binary(command[0]):
            PR.warn("\'{0}\' tool does not exist. ".format(command[0]) +
                    "File system will not be created")
            return 1
        if self.options.force:
            if fstype == 'xfs':
                command.insert(1, '-f')
            if fstype in EXTN:
                command.insert(1, '-F')
        if self.options.verbose:
            if fstype in EXTN:
                command.insert(1, '-v')
        return misc.run(command, stdout=True)[0]

    def _do_mount(self, volume, options=None, directory=None):
        if directory is None:
            directory = self._mpoint
        try:
            volume.mount(directory, options)
        except AttributeError:
            misc.do_mount(volume['real_dev'], directory, options)

    def check(self, args):
        """
        Check the file system on the volume. FsInfo is used for that purpose,
        except for btrfs.
        """
        err = 0
        checked = 0
        for fs in args.device:
            print "Checking {0} file system on \'{1}\':".format(fs.fstype,
                                                                fs.device),
            if fs.mounted:
                print
                try:
                    if PR.check(PR.FS_MOUNTED, [fs.device, fs.mounted]):
                        misc.do_umount(fs.device)
                except problem.FsMounted:
                    PR.warn("Unable to check file system " +
                            "\'{0}\' on volume \'{1}\'".format(fs.fstype,
                                                               fs.device))
                    continue
            ret = fs.fsck()
            checked += 1
            err += ret
            if ret:
                print "FAIL"
            else:
                print "OK"
        if checked == 0:
            PR.error("Nothing was checked")
        if err > 0:
            PR.warn("Some file system(s) contains errors. Please run " +
                    "the appropriate fsck utility")

    def _filter_device_list(self, args, have_size=None, new_size=None):
        """
        Filter the args.device list. Only items which have to be added to
        pool are left in the args.device list. Function returns touple
        (have_size, devices) where have_size is the size of the devices which
        will be added to the pool (args.device) plus optional have_size
        argument. Devices is the list of devices which can be used for volume
        creation, it means that it does not contain devices which are used
        in other pools and are not removed from it in this function.
        """

        if have_size is None:
            have_size = 0.0
        else:
            have_size = float(have_size)

        changed = False

        devices = args.device
        args.device = []

        for dev in devices[:]:
            if self.dev[dev] and 'pool_name' in self.dev[dev] and \
               self.dev[dev]['pool_name'] != args.pool.name:
                if PR.check(PR.DEVICE_USED, [dev, self.dev[dev]['pool_name']]):
                    remove_args = Struct()
                    remove_args.all = False
                    remove_args.items = [self.dev[dev]]
                    if self.remove(remove_args):
                        args.device.append(dev)
                        changed = True
                    elif new_size is None:
                        PR.error("Device \'{0}\' can not be used!".format(dev))
                    else:
                        devices.remove(dev)
                        continue
                # This is tricky. We are going to create or resize a device
                # so we might actually need the device for create (or resize)
                # to finish successfully. Create and resize should check
                # whether is has enough space and fail if it does not. The
                # problem is, when the size was not specified, then the result
                # would be different than what user expected, so we should fail
                # right away.
                elif new_size is None:
                    PR.error("Device \'{0}\' can not be used!".format(dev))
                else:
                    devices.remove(dev)
                    continue

            if not self.dev[dev] or 'pool_name' not in self.dev[dev]:
                # Check signature of existing file system on the device
                # and ask user whether to use it or not.
                signature = misc.get_fs_type(dev)
                if signature and \
                   not PR.check(PR.EXISTING_FILESYSTEM, [signature, dev]):
                    devices.remove(dev)
                    continue
                else:
                    args.device.append(dev)

        if changed:
            self.reinit_dev()

        for dev in devices:
            if not self.dev[dev]:
                have_size += misc.get_file_size(dev)
            else:
                try:
                    have_size += float(self.dev[dev]['dev_free'])
                except ValueError:
                    have_size += float(self.dev[dev]['dev_size'])

        return have_size, devices

    def resize(self, args):
        """
        Resize the volume to the given size. If more devices are provided as
        arguments, it will be added into the pool prior to the volume resize
        only if the space in the pool is not sufficient. That said, only the
        number of devices are added into the pool to be able to cover the
        resize.
        """
        args.pool = self.pool[args.volume['pool_name']]
        vol_size = float(args.volume['vol_size'])

        if not args.size:
            new_size = vol_size
        elif args.size[0] == '+':
            new_size = vol_size + float(args.size[1:])
        elif args.size[0] == '-':
            new_size = vol_size + float(args.size)
        else:
            new_size = float(args.size)
        size_change = new_size - vol_size

        fs = True if 'fs_type' in args.volume else False

        if new_size <= 0:
            PR.error("New volume size \'{0} KB\' is too small".format(new_size))

        if vol_size == new_size:
            # Try to grow the file system, since there is nothing to
            # do with the volume itself.
            if fs:
                args.volume['fs_info'].resize()
            else:
                PR.check(PR.RESIZE_ALREADY_MATCH, [args.volume.name, new_size])
            return

        # Backend might not support pooling
        if args.pool is None:
            pool_free = None
            pool_name = "none"
        else:
            pool_free = float(args.pool['pool_free'])
            pool_name = args.pool.name

        have_size, devices = self._filter_device_list(args,
                                                      pool_free,
                                                      new_size)

        if have_size < size_change:
            PR.check(PR.RESIZE_NOT_ENOUGH_SPACE,
                     [pool_name, args.volume.name, new_size])
        elif len(args.device) > 0 and new_size > vol_size:
            self.add(args, True)

        if new_size != vol_size:
            args.volume.resize(new_size, fs)

    def create(self, args):
        """
        Create new volume (or subvolume in case of btrfs) using the devices
        provided as arguments. If the device is not in the selected pool, then
        add() is called on the pool prior to create().
        """
        # Get the size in kilobytes
        if args.size:
            args.size = misc.get_real_size(args.size)

        if self._mpoint and not (args.fstype or args.pool.type == 'btrfs'):
            if PR.check(PR.CREATE_MOUNT_NOFS, self._mpoint):
                self._mpoint = None

        devices = args.device
        if args.pool.exists():
            pool_free = float(args.pool['pool_free'])
        else:
            pool_free = 0.0

        # If devices were provided we should only use those
        if len(devices) > 0:
            pool_free = None

        have_size, devices = self._filter_device_list(args, pool_free,
                                                      args.size)

        # Currently we do not allow setting subvolume size with btrfs. This
        # should change in the future (quotas maybe) so the check should
        # be removed or pushed to the backend itself.
        if args.size and have_size < float(args.size) and \
           not (args.pool.exists() and args.pool.type == 'btrfs'):
            if PR.check(PR.CREATE_NOT_ENOUGH_SPACE,
                        [have_size, args.pool.name]):
                args.size = None

        # When the pool does not exist and there is no device usable
        # for creating the new pool, then there is no point of trying to
        # create a volume, since it would fail in the backend anyway.
        if not args.pool.exists() and len(devices) == 0:
            PR.check(PR.NO_DEVICES, args.pool.name)

        if have_size == 0:
            PR.error("Not enough space ({0} KB) to".format(have_size) +
                     "to create volume")

        # Number of stripes must not exceed number of devices within the pool
        if args.stripes and len(devices) > 0 and args.stripes > len(devices):
            PR.error("Number of stripes ({0}) ".format(args.stripes) +
                     "must not exceed number of devices " +
                     "({0})".format(len(devices)))
        elif args.stripes and len(devices) == 0 and args.pool.exists():
            tmp = int(args.pool['dev_count'])
            if args.stripes > tmp:
                PR.error("Number of stripes ({0}) ".format(args.stripes) +
                         "must not exceed number of devices " +
                         "({0})".format(tmp))

        # If we want btrfs pool and it does not exist yet, we do not
        # want to call add since it would create it. Note that when
        # btrfs pool is created the new btrfs volume is created as well
        # because it is actually the same thing
        if len(args.device) > 0 and \
           not (not args.pool.exists() and args.pool.type == 'btrfs'):
            self.add(args, True)

        if args.raid:
            raid = {'level': args.raid,
                    'stripesize': args.stripesize,
                    'stripes': args.stripes}
        else:
            raid = None

        lvname = args.pool.create(devs=devices,
                                  size=args.size,
                                  raid=raid,
                                  name=args.name)

        if args.fstype and args.pool.type != 'btrfs':
            if self._create_fs(args.fstype, lvname) != 0:
                self._mpoint = None
        if self._mpoint:
            self.reinit_vol()
            self._do_mount(self.vol[lvname])

    def list(self, args):
        """
        List devices, pools, volumes
        """
        if not args.type:
            self.dev.ptable()
            self.pool.ptable()
            self.vol.ptable(more_data=self.dev.filesystems())
            self.snap.ptable()
        elif args.type in ['fs', 'filesystems']:
            self.vol.ptable(more_data=self.dev.filesystems(), cond="fs_only")
        elif args.type in ['dev', 'devices']:
            self.dev.ptable()
        elif args.type in ["volumes", "vol"]:
            self.vol.ptable(more_data=self.dev.filesystems())
        elif args.type in ["pool", "pools"]:
            self.pool.ptable()
        elif args.type in ['snap', 'snapshots']:
            self.snap.ptable()

    def add(self, args, skip_check=False):
        """
        Add devices into the pool
        """
        if not skip_check:
            for dev in args.device[:]:
                item = self.dev[dev]
                if item and 'pool_name' in item:
                    if item['pool_name'] == args.pool.name:
                        args.device.remove(dev)
                        continue
                    if PR.check(PR.DEVICE_USED,
                                [item.name, item['pool_name']]):
                        remove_args = Struct()
                        remove_args.all = False
                        remove_args.items = [item]
                        if not self.remove(remove_args):
                            args.device.remove(item.name)
                    else:
                        args.device.remove(dev)
                else:
                    # Check signature of existing file system on the device
                    # and as user whether to use it or not.
                    signature = misc.get_fs_type(dev)
                    if signature and \
                       not PR.check(PR.EXISTING_FILESYSTEM, [signature, dev]):
                        args.device.remove(dev)
                        continue

        if args.pool.exists():
            if len(args.device) > 0:
                args.pool.extend(args.device)
            else:
                PR.check(PR.NO_DEVICES, args.pool.name)
        else:
            if len(args.device) > 0:
                args.pool.new(args.device)
            else:
                PR.check(PR.NO_DEVICES, args.pool.name)

    def remove(self, args):
        """
        Remove the all the items, or all pools if all argument is specified.
        Items could be the devices, pools or volumes.
        """
        ret = True
        removed = 0
        if args.all:
            for pool in self.pool:
                try:
                    pool.remove()
                    removed += 1
                except (RuntimeError, problem.SsmError), ex:
                    PR.info("Unable to remove '{0}'".format(pool['pool_name']))
                    ret = False
            if removed == 0:
                PR.error("Nothing was removed!")
            return ret
        if len(args.items) == 0:
            err = "too few arguments"
            raise argparse.ArgumentTypeError(err)
        for item in args.items:
            try:
                if isinstance(item.obj, DeviceInfo):
                    pool = self.pool[item['pool_name']]
                    if pool:
                        pool.reduce(item.name)
                        removed += 1
                        continue
                    else:
                        PR.error("It is not clear what do you want " +
                                 "to achieve by removing " +
                                 "{0}!".format(item.name))
                item.remove()
                removed += 1
            except (RuntimeError, problem.SsmError), ex:
                PR.info("Unable to remove '{0}'".format(item.name))
                ret = False
        if removed == 0:
            PR.error("Nothing was removed")
        return ret

    def snapshot(self, args):
        """
        Create a new snapshot of the volume.
        """
        pool = self.pool[args.volume['pool_name']]
        vol_size = float(args.volume['vol_size'])
        pool_free = float(pool['pool_free'])

        if not args.size:
        # We'll ceate snapshot of the size of 20% of the original volume
            snap_size = vol_size * 0.20
            user_set_size = False
        else:
            snap_size = float(misc.get_real_size(args.size))
            user_set_size = True

        if pool_free < snap_size:
            snap_size = pool_free

        if snap_size <= 0 and pool.type != 'btrfs':
            PR.error("Not enough space ({0} KB) to".format(pool_free) +
                     "to create snapshot")

        args.volume.snapshot(args.dest, args.name, snap_size, user_set_size)

    def mount(self, args):
        """
        Mount a volume at given directory.
        """
        vol = self.vol[args.volume]
        try:
            if vol:
                self._do_mount(vol, args.options, args.directory)
            else:
                misc.do_mount(args.volume, args.directory, args.options)
        except RuntimeError:
            PR.error("Could not mount {0} to ".format(args.volume) +
                     "{0} with options \'{1}\'".format(args.directory,
                                                       args.options))

    def is_fs(self, device):
        real = misc.get_real_device(device)

        vol = self.vol[real]
        if vol and 'fs_type' in vol:
            return vol['fs_info']
        dev = self.dev[real]
        if dev and 'fs_type' in dev:
            return dev['fs_info']
        err = "'{0}' does not contain valid file system".format(real)
        raise argparse.ArgumentTypeError(err)

    def _find_device_record(self, path):
        """
        Try to find device name for path, which is used as an key in
        self.dev - this is usually the real block device, but in some
        rare cases (dmsetup) we can have real block device which name
        does not correspond with what we have in /proc/partitions
        """
        if self.dev[path]:
            return path

        minor = os.minor(os.lstat(path).st_rdev)
        dm_dev = "/dev/dm-{0}".format(minor)
        if self.dev[dm_dev]:
            return dm_dev
        else:
            return path

    def check_create_item(self, path):
        """
        Check the create argument for block device or directory.
        """
        if not self._mpoint:
            try:
                mode = os.stat(path).st_mode
            except OSError:
                err = "'{0}' does not exist.".format(path)
                raise argparse.ArgumentTypeError(err)
            if stat.S_ISDIR(mode):
                self._mpoint = path
                return
        return self.get_bdevice(path)

    def get_bdevice(self, path):
        path = is_bdevice(path)
        return self._find_device_record(path)

    def is_pool(self, string):
        pool = self.pool[string]
        if not pool:
            if string:
                self.pool.default.name = string
            pool = self.pool.default
        return pool

    def is_volume(self, string):
        vol = self.vol[string]
        if vol:
            return vol
        dev = self.dev[string]
        if dev and 'fs_type' in dev:
            return dev
        err = "'{0}' is not a valid volume to resize".format(string)
        raise argparse.ArgumentTypeError(err)

    def can_snapshot(self, string):
        vol = self.vol[string]
        have = False
        if not vol:
            for vol in self.vol:
                if 'mount' in vol and (vol['mount'] == string.rstrip("/")):
                    have = True
                    break
        else:
            have = True
        if not have:
            err = "'{0}' is not valid volume nor mount point.".format(string)
            raise argparse.ArgumentTypeError(err)
        else:
            err = "Backend for '{0}' ".format(string) + \
                  "does not support snapshotting."
            try:
                if not getattr(vol, "snapshot"):
                    raise argparse.ArgumentTypeError(err)
                else:
                    return vol
            except AttributeError:
                raise argparse.ArgumentTypeError(err)

    def check_remove_item(self, string):
        """
        Check the remove argument for volume, pool or device.
        """
        volume = self.vol[string]
        if volume:
            return volume
        pool = self.pool[string]
        if pool:
            return pool
        device = self.dev[string]
        if device:
            return device
        else:
            try:
                path = self.get_bdevice(string)
                device = self.dev[path]
                if device:
                    return device
            except argparse.ArgumentTypeError:
                pass
        for vol in self.vol:
            if 'mount' in vol and (vol['mount'] == string.rstrip("/")):
                return vol
        err = "'{0}' is not valid pool nor volume".format(string)
        raise argparse.ArgumentTypeError(err)


def valid_resize_size(size):
    """
    Validate that the 'size' is usable as resize argument. It means that the
    'size' argument should be in this format: [+|-]number[unit]. It returns the
    number with the provided sign (even with the plus sign) converted to the
    kilobytes. Is no unit is specified, default is kilobytes.

    >>> valid_resize_size("3.14")
    '3.14'
    >>> valid_resize_size("+3.14")
    '+3.14'
    >>> valid_resize_size("-3.14")
    '-3.14'
    >>> valid_resize_size("3.14k")
    '3.14'
    >>> valid_resize_size("+3.14K")
    '+3.14'
    >>> valid_resize_size("-3.14k")
    '-3.14'
    >>> valid_resize_size("3.14G")
    '3292528.64'
    >>> valid_resize_size("+3.14g")
    '+3292528.64'
    >>> valid_resize_size("-3.14G")
    '-3292528.64'
    >>> valid_resize_size("G")
    Traceback (most recent call last):
    ...
    ArgumentTypeError: 'G' is not valid number for the resize.
    """
    try:
        return misc.get_real_size(size)
    except Exception:
        err = "'{0}' is not valid number for the resize.".format(size)
        raise argparse.ArgumentTypeError(err)


def is_directory(self, string):
    try:
        mode = os.stat(string).st_mode
    except OSError:
        err = "'{0}' does not exist.".format(string)
        raise argparse.ArgumentTypeError(err)
    if stat.S_ISDIR(mode):
        return string
    else:
        err = "'{0}' is not directory.".format(string)
        raise argparse.ArgumentTypeError(err)


def is_bdevice(path):
    path = misc.get_real_device(path)
    try:
        mode = os.lstat(path).st_mode
    except OSError:
        err = "'{0}' is not valid block device".format(path)
        raise argparse.ArgumentTypeError(err)
    if not stat.S_ISBLK(mode):
        err = "'{0}' is not valid block device".format(path)
        raise argparse.ArgumentTypeError(err)
    return path


def is_supported_fs(fs):
    if fs in SUPPORTED_FS:
        return fs
    err = "'{0}' is not supported file system".format(fs)
    raise argparse.ArgumentTypeError(err)


class SsmParser(object):
    """
    This class is used to generate argparse parser and run the actual
    parsing.
    """

    def __init__(self, storage, prog=None):
        self.storage = storage
        self.parser = self._get_parser_global(prog)
        self.subcommands = self.parser.add_subparsers(title="Commands")
        self.parser_check = self._get_parser_check()
        self.parser_resize = self._get_parser_resize()
        self.parser_create = self._get_parser_create()
        self.parser_list = self._get_parser_list()
        self.parser_add = self._get_parser_add()
        self.parser_remove = self._get_parser_remove()
        self.parser_snapshot = self._get_parser_snapshot()
        self.parser_mount = self._get_parser_mount()
        self.args = None

    def parse(self):
        self.args = self.parser.parse_args()
        return self.args

    def _get_parser_global(self, prog):
        """
        General ssm options
        """
        parser = argparse.ArgumentParser(
                description="System Storage Manager", prog=prog,
                epilog='''To get help for particular command please specify
                       \'%(prog)s [command] -h\'.''')
        parser.add_argument('--version', action='version',
                version='%(prog)s {0}'.format(VERSION))
        parser.add_argument('-v', '--verbose',
                help="Show aditional information while executing.",
                action="store_true")
        parser.add_argument('-f', '--force',
                help="Force execution in the case where ssm has some " +
                     "doubts or questions.",
                action="store_true")
        parser.add_argument('-b', '--backend', nargs=1,
                metavar='BACKEND',
                help="Choose backend to use. Currently you can choose from " +
                     "({0}).".format(",".join(SUPPORTED_BACKENDS)),
                choices=SUPPORTED_BACKENDS,
                action=SetBackend)
        return parser

    def _get_parser_check(self):
        """
        Check command
        """
        parser_check = self.subcommands.add_parser("check",
                help="Check consistency of the file system on the device.")
        parser_check.add_argument('device', nargs='+',
                help="Device with file system to check.",
                type=self.storage.is_fs)
        parser_check.set_defaults(func=self.storage.check)
        return parser_check

    def _get_parser_resize(self):
        """
        Resize command
        """
        parser_resize = self.subcommands.add_parser("resize",
                help="Change or set the volume and file system size.")
        parser_resize.add_argument("volume", help="Volume to resize.",
                type=self.storage.is_volume)
        parser_resize.add_argument('-s', '--size',
                help='''New size of the volume. With the + or - sign the
                     value is added to or subtracted from the actual size of
                     the volume and without it, the value will be set as the
                     new volume size. A size suffix of [k|K] for kilobytes,
                     [m|M] for megabytes, [g|G] for gigabytes, [t|T] for
                     terabytes or [p|P] for petabytes is optional. If no unit
                     is provided the default is kilobytes.''',
                type=valid_resize_size)
        parser_resize.add_argument("device", nargs='*',
                help='''Devices to use for extending the volume. If the
                     device is not in any pool, it is added into the
                     volume's pool prior to the extension. Note that only
                     really needed number of devices are added into the pool
                     prior the resize.''')
        parser_resize.set_defaults(func=self.storage.resize)
        return parser_resize

    def _get_parser_create(self):
        """
        Create command
        """
        parser_create = self.subcommands.add_parser("create",
                help="Create a new volume with defined parameters.")
        parser_create.add_argument('-s', '--size',
                help='''Gives the size to allocate for the new logical volume
                     A size suffix K|k, M|m, G|g, T|t, P|p, E|e can be used
                     to define 'power of two' units. If no unit is provided, it
                     defaults to kilobytes. This is optional if if
                     not given maximum possible size will be used.''')
        parser_create.add_argument('-n', '--name',
                help='''The name for the new logical volume. This is optional
                     and if omitted, name will be generated by the
                     corresponding backend.''')
        parser_create.add_argument('--fstype',
                help='''Gives the file system type to create on the new
                     logical volume. Supported file systems are (ext3,
                     ext4, xfs, btrfs). This is optional and if not given
                     file system will not be created.''',
                type=is_supported_fs)
        parser_create.add_argument('-r', '--raid', choices=SUPPORTED_RAID,
                metavar="LEVEL",
                help='''Specify a RAID level you want to use when creating a new
                     volume. Note that some backends might not implement all
                     supported RAID levels. This is optional and if no specified,
                     linear volume will be created. You can choose from the
                     following list of supported levels
                     ({0}).'''.format(",".join(SUPPORTED_RAID)))
        parser_create.add_argument('-I', '--stripesize',
                type=int,
                help='''Gives the number of kilobytes for the granularity
                        of stripes. This is optional and if not given, backend
                        default will be used. Note that you have to specify RAID
                        level as well.''')
        parser_create.add_argument('-i', '--stripes',
                type=int,
                help='''Gives the number of stripes. This is equal to the
                     number of physical volumes to scatter the logical
                     volume. This is optional and if stripesize is set
                     and multiple devices are provided stripes is
                     determined automatically from the number of devices. Note
                     that you have to specify RAID level as well.''')
        parser_create.add_argument('-p', '--pool', default="",
                help="Pool to use to create the new volume.",
                type=self.storage.is_pool)
        parser_create.add_argument('device', nargs='*',
                help='''Devices to use for creating the volume. If the device
                     is not in any pool, it is added into the pool prior the
                     volume creation.''',
                type=self.storage.check_create_item,
                action=StoreAll)
        parser_create.add_argument('mount', nargs='?',
                help='''Directory to mount the newly create volume to.''')
        parser_create.set_defaults(func=self.storage.create)
        return parser_create

    def _get_parser_list(self):
        """
        List command
        """
        parser_list = self.subcommands.add_parser("list",
                help='''List information about
                     all detected, devices, pools, volumes and snapshots
                     in the system.''')
        parser_list.add_argument('type', nargs='?',
                choices=["volumes", "vol", "dev", "devices", "pool", "pools",
                    "fs", "filesystems", "snap", "snapshots"])
        parser_list.set_defaults(func=self.storage.list)
        return parser_list

    def _get_parser_add(self):
        """
        Add command
        """
        parser_add = self.subcommands.add_parser("add",
                help='''Add one or more devices into the pool.''')
        parser_add.add_argument('-p', '--pool', default="",
                help='''Pool to add device into. If not specified the default
                     pool is used.''', type=self.storage.is_pool)
        parser_add.add_argument('device', nargs='+',
                help="Devices to add into the pool.",
                type=self.storage.get_bdevice,
                action=StoreAll)
        parser_add.set_defaults(func=self.storage.add)
        return parser_add

    def _get_parser_remove(self):
        """
        Remove command
        """
        parser_remove = self.subcommands.add_parser("remove",
                help='''Remove devices from the pool, volumes or pools.''')
        parser_remove.add_argument('-a', '--all', action="store_true",
                help="Remove all pools in the system.")
        parser_remove.add_argument('items', nargs='*',
                help="Items to remove. Item could be device, pool, or volume.",
                type=self.storage.check_remove_item)
        parser_remove.set_defaults(func=self.storage.remove)
        return parser_remove

    def _get_parser_snapshot(self):
        """
        Snapshot command
        """
        parser_snapshot = self.subcommands.add_parser("snapshot",
                help='''Take a snapshot of the existing volume.''')
        parser_snapshot.add_argument('-s', '--size',
                help='''Gives the size to allocate for the new snapshot volume
                     A size suffix K|k, M|m, G|g, T|t, P|p, E|e can be used
                     to define 'power of two' units. If no unit is provided, it
                     defaults to kilobytes. This is option and if not give,
                     the size will be determined automatically.''')
        group = parser_snapshot.add_mutually_exclusive_group()
        group.add_argument('-d', '--dest',
                help='''Destination of the snapshot specified with absolute
                     path to be used for the new snapshot. This is optional
                     and if not specified default backend policy will be
                     performed.''')
        group.add_argument('-n', '--name',
                help='''Name of the new snapshot. This is optional and if not
                     specified  default backend policy will be performed.''')

        parser_snapshot.add_argument('volume',
                help="Volume, or mount point to take a snapshot of.",
                type=self.storage.can_snapshot)
        parser_snapshot.set_defaults(func=self.storage.snapshot)
        return parser_snapshot

    def _get_parser_mount(self):
        """
        Mount command
        """
        parser_mount = self.subcommands.add_parser("mount",
                help='''Mount a volume with file system to specified
                     locaion.''')
        parser_mount.add_argument('-o', '--options',
                help='''Options are specified with a -o flag followed by a
                     comma separated string of options. This option is
                     equivalent to the same mount(8) option.''')
        parser_mount.add_argument("volume", help="Volume to mount.")
        parser_mount.add_argument("directory",
                help="Directory to attach the volume.",
                type=is_directory)
        parser_mount.set_defaults(func=self.storage.mount)
        return parser_mount


def main(args=None):

    if args:
        sys.argv = args.split()

    options = Options()
    PR.set_options(options)
    storage = StorageHandle(options)
    ssm_parser = SsmParser(storage)
    args = ssm_parser.parse()

    # Check create command dependency
    if args.func == storage.create:
        if not args.raid:
            if (args.stripesize):
                err = "You can not specify --stripesize without specifying" + \
                      " RAID level!"
                ssm_parser.parser_create.error(err)
            if (args.stripes):
                err = "You can not specify --stripes without specifying" + \
                      " RAID level!"
                ssm_parser.parser_create.error(err)

    options.verbose = args.verbose
    options.force = args.force

    #storage.set_globals(args.force, args.verbose, args.yes, args.config)
    storage.set_globals(options)

    # Register clean-up function on exit
    sys.exitfunc = misc.do_cleanup

    try:
        args.func(args)
    except argparse.ArgumentTypeError, ex:
        ssm_parser.parser.error(ex)

    return 0



#!/usr/bin/env python
#
# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Unittests for the system storage manager

import os
import re
import sys
import stat
import time
import doctest
import unittest
import argparse
from ssmlib import main
from ssmlib import misc
from ssmlib import problem

from tests.unittests.common import *

class SimpleStorageHandleSanityCheck(BaseStorageHandleInit):
    """
    Simple sanity ckecks for StorageHandle class and some of its methods.
    """

    def test_constructor(self):
        # Check initial variables
        self.assertFalse(self.storage.options.force)
        self.assertFalse(self.storage.options.verbose)
        self.assertFalse(self.storage.options.yes)
        self.assertFalse(self.storage.options.interactive)
        self.assertFalse(self.storage.options.debug)
        self.assert_(self.storage.options.config is None)
        self.assert_(self.storage._mpoint is None)
        self.assert_(self.storage._dev is None)
        self.assert_(self.storage._pool is None)
        self.assert_(self.storage._volumes is None)

    def test_create_fs(self):
        options = main.Options()
        for fs in main.EXTN:
            options.force = True
            options.verbose = True
            options.yes = False
            options.config = "my_new_config"
            options.interactive = False
            self.storage.set_globals(options)
            self.storage._create_fs(fs, "/dev/foo/bar")
            self.assertEqual('mkfs.{0} -v -F /dev/foo/bar'.format(fs),
                    self.run_data[-1])
            options.force = False
            options.verbose = False
            self.storage.set_globals(options)
            self.storage._create_fs(fs, "/dev/foo/bar")
            self.assertEqual('mkfs.{0} /dev/foo/bar'.format(fs),
                    self.run_data[-1])
            options.force = True
            self.storage.set_globals(options)
            self.storage._create_fs(fs, "/dev/foo/bar")
            self.assertEqual('mkfs.{0} -F /dev/foo/bar'.format(fs),
                    self.run_data[-1])
            options.force = False
            options.verbose = True
            self.storage.set_globals(options)
            self.storage._create_fs(fs, "/dev/foo/bar")
            self.assertEqual('mkfs.{0} -v /dev/foo/bar'.format(fs),
                    self.run_data[-1])
        options.force = True
        options.verbose = True
        self.storage.set_globals(options)
        self.storage._create_fs("xfs", "/dev/foo/bar")
        self.assertEqual('mkfs.xfs -f /dev/foo/bar', self.run_data[-1])
        options.force = False
        options.verbose = False
        self.storage.set_globals(options)
        self.storage._create_fs("xfs", "/dev/foo/bar")
        self.assertEqual('mkfs.xfs /dev/foo/bar', self.run_data[-1])
        options.force = True
        self.storage.set_globals(options)
        self.storage._create_fs("xfs", "/dev/foo/bar")
        self.assertEqual('mkfs.xfs -f /dev/foo/bar', self.run_data[-1])
        options.force = False
        options.verbose = True
        self.storage.set_globals(options)
        self.storage._create_fs("xfs", "/dev/foo/bar")
        self.assertEqual('mkfs.xfs /dev/foo/bar', self.run_data[-1])


class StorageHandleSanityCheck(BaseStorageHandleInit):
    """
    Sanity checks for StorageHandle class and other classes used via
    StorageHandle class.
    """

    def setUp(self):
        super(StorageHandleSanityCheck, self).setUp()
        options = main.Options()
        options.force = True
        options.verbose = True
        options.yes = True
        options.config = "my_config"
        options.interactive = True
        options.debug = True
        self.storage.set_globals(options)
        self.dev = self.storage.dev
        self.vol = self.storage.vol
        self.pool = self.storage.pool

    def tearDown(self):
        super(StorageHandleSanityCheck, self).tearDown()
        self.dev = None
        self.vol = None
        self.pool = None

    def test_storage_constructor(self):
        # Check initial variables
        self.assertTrue(self.storage.options.force)
        self.assertTrue(self.storage.options.verbose)
        self.assertTrue(self.storage.options.yes)
        self.assertTrue(self.storage.options.interactive)
        self.assertTrue(self.storage.options.debug)
        self.assertEqual(self.storage.options.config, "my_config")

        # Check if we have right instances
        self.assert_(isinstance(self.dev, main.Devices))
        self.assert_(isinstance(self.vol, main.Volumes))
        self.assert_(isinstance(self.pool, main.Pool))

        # Check initial variables
        for source in self.dev, self.vol, self.pool:
            self.assertTrue(source.options.force)
            self.assertTrue(source.options.verbose)
            self.assertTrue(source.options.yes)
            self.assertTrue(source.options.interactive)
            self.assertTrue(source.options.debug)
            self.assertEqual(source.options.config, "my_config")
            self.assert_(isinstance(source._data, dict))
            self.assert_(len(source._data) >  0)
            self.assert_(isinstance(source.header, list))
            self.assert_(len(source.header) >  0)
            self.assert_(isinstance(source.attrs, list))
            self.assert_(len(source.attrs) >  0)
            self.assert_(isinstance(source.types, list))
            self.assert_(len(source.types) >  0)
            for item  in source._data.itervalues():
                self.assertTrue(item.options.force)
                self.assertTrue(item.options.verbose)
                self.assertTrue(item.options.yes)
                self.assertTrue(item.options.interactive)
                self.assertTrue(item.options.debug)
                self.assertEqual(item.options.config, "my_config")

    def test_set_globals_propagation(self):
        options = main.Options()
        options.force = False
        options.verbose = False
        options.yes = False
        options.config = "my_config"
        options.interactive = False
        options.debug = False
        self.storage.set_globals(options)
        # Check initial variables
        self.assertFalse(self.storage.options.force)
        self.assertFalse(self.storage.options.verbose)
        self.assertFalse(self.storage.options.yes)
        self.assertEqual(self.storage.options.config, "my_config")
        self.assertFalse(self.storage.options.interactive)
        self.assertFalse(self.storage.options.debug)

        # Check initial variables
        for source in self.dev, self.vol, self.pool:
            self.assertFalse(source.options.force)
            self.assertFalse(source.options.verbose)
            self.assertFalse(source.options.yes)
            self.assertFalse(source.options.interactive)
            self.assertFalse(source.options.debug)
            self.assertEqual(source.options.config, "my_config")
            self.assert_(isinstance(source._data, dict))
            self.assert_(len(source._data) > 0)
            self.assert_(isinstance(source.header, list))
            self.assert_(len(source.header) > 0)
            self.assert_(isinstance(source.attrs, list))
            self.assert_(len(source.attrs) > 0)
            self.assert_(isinstance(source.types, list))
            self.assert_(len(source.types) > 0)
            for item  in source._data.itervalues():
                self.assertFalse(item.options.force)
                self.assertFalse(item.options.verbose)
                self.assertFalse(item.options.yes)
                self.assertFalse(item.options.interactive)
                self.assertFalse(item.options.debug)
                self.assertEqual(item.options.config, "my_config")

    def test_backend_generic_methods(self):
        for source in self.dev, self.vol, self.pool:
            obj = dir(source)
            self.assert_("__iter__" in obj)
            self.assert_("__contains__" in obj)
            self.assert_("__getitem__" in obj)
            self.assert_("filesystems" in obj)
            self.assert_("ptable" in obj)
            self.assert_("set_globals" in obj)
            # Variables
            self.assert_("_data" in obj)
            self.assert_("header" in obj)
            self.assert_("attrs" in obj)
            self.assert_("types" in obj)
            for bknd in source._data.itervalues():
                obj = dir(bknd)
                self.assert_("__iter__" in obj)
                self.assert_("__getitem__" in obj)
                self.assert_("__init__" in obj)
                # Variables
                self.assert_("type" in obj)
                self.assert_("options" in obj)
                # DeviceInfo does not need default_pool_name
                if bknd.type != "device":
                    self.assert_("default_pool_name" in obj)

    def test_volumes_specific_methods(self):
        for bknd in self.vol._data.itervalues():
            obj = dir(bknd)
            self.assert_("remove" in obj)

    def test_pool_specific_methods(self):
        for bknd in self.pool._data.itervalues():
            obj = dir(bknd)
            self.assert_("reduce" in obj)
            self.assert_("remove" in obj)
            self.assert_("create" in obj)
            self.assert_("new" in obj)
            self.assert_("extend" in obj)


class SimpleSsmSanityCheck(unittest.TestCase):
    """
    Simple sanity check of ssm module.
    """

    def test_existing_objects(self):
        obj = dir(main)
        self.assert_("StorageHandle" in obj)
        self.assert_("FsInfo" in obj)
        self.assert_("DeviceInfo" in obj)
        self.assert_("SUPPORTED_FS" in obj)
        self.assert_("DEFAULT_DEVICE_POOL" in obj)
        self.assert_("Storage" in obj)
        self.assert_("Pool" in obj)
        self.assert_("Devices" in obj)
        self.assert_("Volumes" in obj)

    def test_fsinfo_methods(self):
        obj = dir(main.FsInfo)
        self.assert_("extN_get_info" in obj)
        self.assert_("extN_fsck" in obj)
        for fs in set(main.SUPPORTED_FS) - set(main.EXTN):
            if fs == 'btrfs':
                continue
            self.assert_("{0}_get_info".format(fs) in obj)
            self.assert_("{0}_fsck".format(fs) in obj)

    def test_storage_handle_methods(self):
        obj = dir(main.StorageHandle)
        self.assert_("dev" in obj)
        self.assert_("pool" in obj)
        self.assert_("vol" in obj)
        self.assert_("check" in obj)
        self.assert_("resize" in obj)
        self.assert_("create" in obj)
        self.assert_("list" in obj)
        self.assert_("add" in obj)
        self.assert_("remove" in obj)
        self.assert_("snapshot" in obj)
        self.assert_("mount" in obj)
        self.assert_("set_globals" in obj)


class SsmFunctionCheck(MockSystemDataSource):
    """
    Prepare data for ssm function check.
    """

    def setUp(self):
        super(SsmFunctionCheck, self).setUp()
        self.volumes_orig = main.Volumes
        main.Volumes = Volumes
        self.pool_orig = main.Pool
        main.Pool = Pool
        self.devices_orig = main.Devices
        main.Devices = Devices
        self._addDevice('/dev/sda', 11489037516)
        self._addDevice('/dev/sdb', 234566451)
        self._addDevice('/dev/sdc', 2684354560)
        self._addDevice('/dev/sdc1', 894784853, 1)
        self._addDevice('/dev/sdc2', 29826161, 2)
        self._addDevice('/dev/sdc3', 1042177280, 3)
        self._addDevice('/dev/sdd', 11673)

    def tearDown(self):
        super(SsmFunctionCheck, self).tearDown()
        main.Volumes = self.volumes_orig
        main.Pool = self.pool_orig
        main.Devices = self.devices_orig

    def mock_run(self, cmd, *args, **kwargs):
        for i,arg in enumerate(cmd):
            if arg is None:
                cmd[i] = ''
            elif type(arg) is not str:
                cmd[i] = str(arg)

        self.run_data.append(re.sub('\s\s+',' '," ".join(cmd).strip()))
        output = ""
        if cmd[0] == 'pooldata':
            return self.pool_data
        elif cmd[0] == 'volumedata':
            return self.vol_data
        elif cmd[0] == 'devdata':
            return self.dev_data
        if 'return_stdout' in kwargs and not kwargs['return_stdout']:
            output = None
        return (0, output)

    def test_resize(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3'])
        self._addVol('vol001', 2982616, 1, 'my_pool', ['/dev/sdc2'],
                    '/mnt/test1')
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addDevice('/dev/sde', 11489037516)

        # Extend Volume
        self._checkCmd("ssm resize", ['--size +4m', '/dev/default_pool/vol003'],
            "vol resize /dev/default_pool/vol003 5120.0 False")

        # Shrink volume
        self._checkCmd("ssm resize", ['-s-100G', '/dev/default_pool/vol002'],
            "vol resize /dev/default_pool/vol002 132426625.0 False")

        # Set volume size
        self._checkCmd("ssm resize", ['-s 10M', '/dev/my_pool/vol001'],
            "vol resize /dev/my_pool/vol001 10240.0 False")

        # Set volume and add devices
        self._checkCmd("ssm resize -s 12T /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "vol resize /dev/default_pool/vol003 12884901888.0 False")
        self.assertEqual(self.run_data[-2],
            "pool extend default_pool /dev/sdc1 /dev/sde")

        # Set volume size
        self._checkCmd("ssm resize -s 10G /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "vol resize /dev/default_pool/vol003 10485760.0 False")
        self.assertEqual(self.run_data[-2],
            "pool extend default_pool /dev/sdc1 /dev/sde")

        # Extend volume size with adding more devices
        self._checkCmd("ssm resize -s +10G /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "vol resize /dev/default_pool/vol003 10486784.0 False")
        self.assertEqual(self.run_data[-2],
            "pool extend default_pool /dev/sdc1 /dev/sde")

        # Shrink volume with devices provided
        self._checkCmd("ssm resize -s-10G /dev/default_pool/vol002 /dev/sdc1 /dev/sde",
            [], "vol resize /dev/default_pool/vol002 226798465.0 False")
        self.assertNotEqual(self.run_data[-2],
            "pool extend default_pool /dev/sdc1 /dev/sde")
        self.assertNotEqual(self.run_data[-2],
            "pool extend default_pool /dev/sdc1")

        # Test that we do not use devices which are already used in different
        # pool
        self.assertRaises(Exception, main.main, "ssm resize -s +1.5T /dev/my_pool/vol001 /dev/sdb /dev/sda")

        # If the device we are able to use can cover the size, then
        # it will be resized successfully
        self._checkCmd("ssm resize -s +1.5T /dev/my_pool/vol001 /dev/sdb /dev/sda /dev/sdc1",
            [], "vol resize /dev/my_pool/vol001 1613595352.0 False")

    def test_create(self):
        # Create volume using single device from non existent default pool
        self._checkCmd("ssm create", ['/dev/sda'],
            "pool create {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        self._checkCmd("ssm create", ['--name myvolume', '--fstype ext4', '/dev/sda'])
        self._cmdEq("mkfs.ext4 /dev/{0}/myvolume".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool create {0} myvolume /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)
        self._cmdEq("pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -3)

        self._checkCmd("ssm -f create", ['--fstype ext4', '/dev/sda'])
        self._cmdEq("mkfs.ext4 -F /dev/{0}/lvol001".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("force pool create {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)
        self._cmdEq("force pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -3)

        self._checkCmd("ssm -v create", ['--name myvolume', '--fstype xfs', '/dev/sda'])
        self._cmdEq("mkfs.xfs /dev/{0}/myvolume".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("verbose pool create {0} myvolume /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)
        self._cmdEq("verbose pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -3)

        self._checkCmd("ssm -v -f create", ['--name myvolume', '--fstype xfs', '/dev/sda'])
        self._cmdEq("mkfs.xfs -f /dev/{0}/myvolume".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("force verbose pool create {0} myvolume /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)
        self._cmdEq("force verbose pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -3)

        self._checkCmd("ssm create", ['-s 2.6T', '/dev/sda'],
            "pool create {0} 2791728742.40 /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '-I 16', '/dev/sda'],
            "pool create {0} 2791728742.40 0 16 /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        # Number of stripes must not exceed number of devices
        self.assertRaises(problem.GeneralError, main.main, "ssm create -r 1 -s 2.6T -I 16 -i 4 /dev/sda")

        self._checkCmd("ssm create", ['-r 1', '-s 2.6T', '-I 16', '/dev/sda'],
            "pool create {0} 2791728742.40 1 16 /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        # Create volume using single device from non existent my_pool
        self._checkCmd("ssm create", ['--pool my_pool', '/dev/sda'],
            "pool create my_pool /dev/sda")
        self._cmdEq("pool new my_pool /dev/sda", -2)

        self._checkCmd("ssm create", ['--pool my_pool', '-s 2.6T', '/dev/sda'],
            "pool create my_pool 2791728742.40 /dev/sda")
        self._cmdEq("pool new my_pool /dev/sda", -2)

        self._checkCmd("ssm create", ['-r 10', '-p my_pool', '-s 2.6T', '-I 16',
            '/dev/sda'], "pool create my_pool 2791728742.40 10 16 /dev/sda")
        self._cmdEq("pool new my_pool /dev/sda", -2)

        self._checkCmd("ssm create", ['-r 0', '-p my_pool', '-s 2.6T', '-I 16',
            '/dev/sda'], "pool create my_pool 2791728742.40 0 16 /dev/sda")
        self._cmdEq("pool new my_pool /dev/sda", -2)

        # Create volume using multiple devices
        self._checkCmd("ssm create", ['/dev/sda /dev/sdc1'],
            "pool create {0} /dev/sda /dev/sdc1".format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool new {0} /dev/sda /dev/sdc1".format(main.DEFAULT_DEVICE_POOL), -2)

        self._checkCmd("ssm create", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "pool create my_pool /dev/sda /dev/sdc1")
        self._cmdEq("pool new my_pool /dev/sda /dev/sdc1", -2)

        # Create volume using single device from existing pool
        self._addPool(main.DEFAULT_DEVICE_POOL, ['/dev/sdb', '/dev/sdd'])
        self._checkCmd("ssm create", ['-r 10', '-s 2.6T', '-I 16',
            '-n myvolume', '/dev/sda'],
            "pool create {0} 2791728742.40 myvolume 10 16 /dev/sda". format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool extend {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        self._checkCmd("ssm create", ['-r 1', '-p my_pool', '-s 2.6T', '-I 16',
            '-n myvolume', '/dev/sda'],
            "pool create my_pool 2791728742.40 myvolume 1 16 /dev/sda")
        self._cmdEq("pool extend my_pool /dev/sda", -2)

        # Create volume using multiple devices which one of the is in already
        # in the pool
        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '-I 16',
            '-i 2', '-n myvolume', '/dev/sda /dev/sdb'],
            "pool create {0} 2791728742.40 myvolume 0 2 16 /dev/sda /dev/sdb". format(main.DEFAULT_DEVICE_POOL))
        self._cmdEq("pool extend {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL), -2)

        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        self._checkCmd("ssm create", ['-r 10', '-p my_pool', '-s 2.6T', '-I 16',
            '-i 2', '-n myvolume', '/dev/sdc2 /dev/sda'],
            "pool create my_pool 2791728742.40 myvolume 10 2 16 /dev/sdc2 /dev/sda")
        self._cmdEq("pool extend my_pool /dev/sda", -2)

        # Test that we do not use devices which are already used in different
        # pool
        self.assertRaises(Exception, main.main, "ssm create -p new_pool /dev/sdc2 /dev/sdc3")
        self.assertRaises(Exception, main.main, "ssm create -p new_pool /dev/sdc2 /dev/sdc3 /dev/sda")
        # If the device we are able to use can cover the size, then
        # it will be created
        self._checkCmd("ssm create", ['-s 100M', '-p new_pool', '/dev/sdc2 /dev/sdc3 /dev/sda'],
            "pool create new_pool 102400.00 /dev/sda")

        #sys.stdout = out.stdout

    def test_add(self):
        # Adding to non existent pool
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "pool new {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "pool new {0} /dev/sda /dev/sdc1".format(main.DEFAULT_DEVICE_POOL))
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "pool new my_pool /dev/sda")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "pool new my_pool /dev/sda")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "pool new my_pool /dev/sda /dev/sdc1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "pool new my_pool /dev/sda /dev/sdc1")
        # Add force
        self._checkCmd("ssm -f add", ['--pool my_pool', '/dev/sda'],
            "force pool new my_pool /dev/sda")
        # Add verbose
        self._checkCmd("ssm -v add", ['--pool my_pool', '/dev/sda'],
            "verbose pool new my_pool /dev/sda")
        # Add force and verbose
        self._checkCmd("ssm -v -f add", ['--pool my_pool', '/dev/sda'],
            "force verbose pool new my_pool /dev/sda")

        # Adding to existing default pool
        self._addPool(main.DEFAULT_DEVICE_POOL, ['/dev/sdb', '/dev/sdd'])
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "pool extend {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL))
        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "pool extend {0} /dev/sda /dev/sdc1".format(main.DEFAULT_DEVICE_POOL))

        # Adding to existing defined pool
        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "pool extend my_pool /dev/sda")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "pool extend my_pool /dev/sda")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "pool extend my_pool /dev/sda /dev/sdc1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "pool extend my_pool /dev/sda /dev/sdc1")
        # Add force
        self._checkCmd("ssm -f add", ['--pool my_pool', '/dev/sda'],
            "force pool extend my_pool /dev/sda")
        # Add verbose
        self._checkCmd("ssm -v add", ['--pool my_pool', '/dev/sda'],
            "verbose pool extend my_pool /dev/sda")
        # Add force and verbose
        self._checkCmd("ssm -v -f add", ['--pool my_pool', '/dev/sda'],
            "force verbose pool extend my_pool /dev/sda")

        # Add two devices into existing pool (one of the devices already is in
        # the pool
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sdc2 /dev/sda'],
            "pool extend my_pool /dev/sda")
        self._checkCmd("ssm add", ['/dev/sda /dev/sdb'],
            "pool extend {0} /dev/sda".format(main.DEFAULT_DEVICE_POOL))

    def test_remove(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'])

        # remove volume
        main.main("ssm remove /dev/default_pool/vol002")
        self._cmdEq("vol remove /dev/default_pool/vol002")
        # remove multiple volumes
        main.main("ssm remove /dev/default_pool/vol002 /dev/default_pool/vol003")
        self._cmdEq("vol remove /dev/default_pool/vol002", -2)
        self._cmdEq("vol remove /dev/default_pool/vol003")
        # remove pool
        main.main("ssm remove my_pool")
        self._cmdEq("pool remove my_pool")
        # remove multiple pools
        main.main("ssm remove my_pool default_pool")
        self._cmdEq("pool remove my_pool", -2)
        self._cmdEq("pool remove default_pool")
        # remove device
        main.main("ssm remove /dev/sdc1")
        self._cmdEq("pool reduce my_pool /dev/sdc1")
        # remove multiple devices
        main.main("ssm remove /dev/sdc1 /dev/sdb")
        self._cmdEq("pool reduce my_pool /dev/sdc1", -2)
        self._cmdEq("pool reduce default_pool /dev/sdb")
        # remove combination
        main.main("ssm remove /dev/sdb my_pool /dev/default_pool/vol001")
        self._cmdEq("pool reduce default_pool /dev/sdb", -3)
        self._cmdEq("pool remove my_pool", -2)
        self._cmdEq("vol remove /dev/default_pool/vol001")
        # remove all
        main.main("ssm remove --all")
        self._cmdEq("pool remove default_pool", -2)
        self._cmdEq("pool remove my_pool")
        # remove force
        main.main("ssm -f remove /dev/default_pool/vol002")
        self._cmdEq("force vol remove /dev/default_pool/vol002")
        # remove verbose
        main.main("ssm -v remove /dev/default_pool/vol002")
        self._cmdEq("verbose vol remove /dev/default_pool/vol002")
        # remove verbose + force
        main.main("ssm -v -f remove /dev/default_pool/vol002")
        self._cmdEq("force verbose vol remove /dev/default_pool/vol002")

    def test_snapshot(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/test1')
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'], '/mnt/test')

        # Create snapshot
        self._checkCmd("ssm snapshot", ['/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 41943040.0 False")
        # Create snapshot verbose
        self._checkCmd("ssm -v snapshot", ['/dev/default_pool/vol004'],
            "verbose vol snapshot /dev/default_pool/vol004 41943040.0 False")
        # Create snapshot force
        self._checkCmd("ssm -f snapshot", ['/dev/default_pool/vol004'],
            "force vol snapshot /dev/default_pool/vol004 41943040.0 False")
        # Create snapshot force verbose
        self._checkCmd("ssm -f -v snapshot", ['/dev/default_pool/vol004'],
            "force verbose vol snapshot /dev/default_pool/vol004 41943040.0 False")

        # Create snapshot with size specified
        self._checkCmd("ssm snapshot --size 1G", ['/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 1048576.0 True")
        # Create snapshot with destination specified
        self._checkCmd("ssm snapshot --dest /mnt/test", ['/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 /mnt/test 41943040.0 False")
        # Create snapshot with the name specified
        self._checkCmd("ssm snapshot --name test", ['/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 test 41943040.0 False")
        # Create snapshot with both destination and size specified
        self._checkCmd("ssm snapshot",
            ['--size 1G', '--dest /mnt/test' ,'/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 /mnt/test 1048576.0 True")
        # Create snapshot with both name and size specified
        self._checkCmd("ssm snapshot",
            ['--size 1G', '--name test' ,'/dev/default_pool/vol004'],
            "vol snapshot /dev/default_pool/vol004 test 1048576.0 True")

        # Repeat the test with specifying mount point instead of volume

        # Create snapshot
        self._checkCmd("ssm snapshot", ['/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 41943040.0 False")
        # Create snapshot with size specified
        self._checkCmd("ssm snapshot --size 1G", ['/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 1048576.0 True")
        # Create snapshot with destination specified
        self._checkCmd("ssm snapshot --dest /mnt/test", ['/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 /mnt/test 41943040.0 False")
        # Create snapshot with the name specified
        self._checkCmd("ssm snapshot --name test", ['/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 test 41943040.0 False")
        # Create snapshot with both destination and size specified
        self._checkCmd("ssm snapshot",
            ['--size 1G', '--dest /mnt/test' ,'/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 /mnt/test 1048576.0 True")
        # Create snapshot with both name and size specified
        self._checkCmd("ssm snapshot",
            ['--size 1G', '--name test' ,'/mnt/test'],
            "vol snapshot /dev/default_pool/vol004 test 1048576.0 True")

    def test_mount(self):
        self._addDir("/mnt/test")
        self._addDir("/mnt/test1")
        self._addDir("/mnt/test2")
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/test1')
        self._addVol('vol002', 237284225, 1, 'my_pool', ['/dev/sda'])

        # Simple mount test
        main.main("ssm mount /dev/default_pool/vol001 /mnt/test")
        self._cmdEq("mount /dev/default_pool/vol001 /mnt/test")
        main.main("ssm mount -o discard /dev/default_pool/vol001 /mnt/test")
        self._cmdEq("mount -o discard /dev/default_pool/vol001 /mnt/test")
        main.main("ssm mount --options rw,discard,neco=44 /dev/my_pool/vol002 /mnt/test1")
        self._cmdEq("mount -o rw,discard,neco=44 /dev/my_pool/vol002 /mnt/test1")

        # Non existing volume
        main.main("ssm mount nonexisting /mnt/test1")
        self._cmdEq("mount nonexisting /mnt/test1")
        main.main("ssm mount -o discard,rw nonexisting /mnt/test1")
        self._cmdEq("mount -o discard,rw nonexisting /mnt/test1")


class MyInfo(object):
    def __init__(self, options, data=None):
        self.data = data or {}
        self.options = options
        self.type = 'backend'

    @property
    def y(self):
        return 'yes' if self.options.yes else ''

    @property
    def f(self):
        return 'force' if self.options.force else ''

    @property
    def v(self):
        return 'verbose' if self.options.verbose else ''

    def __iter__(self):
        for item in sorted(self.data.iterkeys()):
            yield item

    def __getitem__(self, key):
        if key in self.data.iterkeys():
            return self.data[key]


class PoolInfo(MyInfo):
    def __init__(self, *args, **kwargs):
        super(PoolInfo, self).__init__(*args, **kwargs)
        self.exist = False
        self.data.update(misc.run(['pooldata']))

    def reduce(self, pool, devices):
        if type(devices) is not list:
            devices = [devices]
        cmd = [self.f, self.v, self.y, 'pool reduce', pool]
        cmd.extend(devices)
        misc.run(cmd)

    def remove(self, pool):
        misc.run([self.f, self.v, self.y, 'pool remove', pool])

    def new(self, pool, devices):
        if type(devices) is not list:
            devices = [devices]
        cmd = [self.f, self.v, self.y, 'pool new', pool]
        cmd.extend(devices)
        misc.run(cmd)

    def extend(self, pool, devices):
        if type(devices) is not list:
            devices = [devices]
        cmd = [self.f, self.v, self.y, 'pool extend', pool]
        cmd.extend(devices)
        misc.run(cmd)

    def create(self, pool, size='', name='', devs='',
               raid=None):
        if type(devs) is not list:
            devices = [devs]
        if raid:
            stripes = raid['stripes']
            stripesize = raid['stripesize']
            level = raid['level']
        else:
            stripes = stripesize = level = ""
        misc.run([self.f, self.v, self.y, 'pool create', pool, size, name,
            level, stripes, stripesize, " ".join(devs)])
        if not name:
            name = "lvol001"
        return "/dev/{0}/{1}".format(pool, name)


class VolumeInfo(MyInfo):
    def __init__(self, *args, **kwargs):
        super(VolumeInfo, self).__init__(*args, **kwargs)
        self.data.update(misc.run(['volumedata']))

    def remove(self, volume):
        misc.run([self.f, self.v, self.y, 'vol remove', volume])

    def resize(self, lv, size, resize_fs=True):
        misc.run([self.f, self.v, self.y, 'vol resize', lv, str(size),
                  str(resize_fs)])

    def snapshot(self, volume, destination, name, size, user_set_size):
        misc.run([self.f, self.v, self.y, 'vol snapshot', volume, destination,
                name, str(size), str(user_set_size)])


class DevInfo(MyInfo):
    def __init__(self, *args, **kwargs):
        super(DevInfo, self).__init__(*args, **kwargs)
        #if len(misc.run(['devdata']) > 0):
        self.data.update(misc.run(['devdata']))

    def remove(self, devices):
        if type(devices) is not list:
            devices = [devices]
        misc.run([self.f, self.v, self.y, 'dev remove', devices])
        cmd.extend(devices)
        misc.run(cmd)


class Pool(main.Storage):
    def __init__(self, *args, **kwargs):
        super(Pool, self).__init__(*args, **kwargs)
        _default_backend = PoolInfo(options=self.options)
        self._data = {'test': _default_backend}
        self.default = main.Item(_default_backend, main.DEFAULT_DEVICE_POOL)
        self.header = ['Pool', 'Devices', 'Free', 'Used', 'Total']
        self.attrs = ['pool_name', 'dev_count', 'pool_free', 'pool_used', 'pool_size']
        self.types = [str, str, float, float, float]


class Volumes(main.Storage):
    def __init__(self, *args, **kwargs):
        super(Volumes, self).__init__(*args, **kwargs)
        self._data = {'test': VolumeInfo(options=self.options)}
        self.header = ['Volume', 'Volume size', 'FS', 'Free',
                       'Used', 'FS size', 'Type', 'Mount point']
        self.attrs = ['dev_name', 'dev_size', 'fs_type',
                      'fs_free', 'fs_used', 'fs_size', 'type', 'mount']
        self.types = [str, float, str, float, float, float, str, str]


class Devices(main.Storage):
    def __init__(self, *args, **kwargs):
        super(Devices, self).__init__(*args, **kwargs)
        self._data = {'dev': main.DeviceInfo(options=self.options,
                                data=DevInfo(options=self.options).data)}
        self.header = ['Device', 'Free', 'Used',
                       'Total', 'Pool', 'Mount point']
        self.attrs = ['dev_name', 'dev_free', 'dev_used', 'dev_size',
                      'pool_name', 'mount']
        self.types = [str, float, float, float, str, str]



#!/usr/bin/env python
#
# (C)2012 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Unittests for the system storage manager btrfs backend


import unittest
from ssmlib import main
from ssmlib.backends import btrfs
from tests.unittests.common import *

class BtrfsFunctionCheck(MockSystemDataSource):

    def setUp(self):
        super(BtrfsFunctionCheck, self).setUp()
        self._addDevice('/dev/sda', 11489037516)
        self._addDevice('/dev/sdb', 234566451)
        self._addDevice('/dev/sdc', 2684354560)
        self._addDevice('/dev/sdc1', 894784853, 1)
        self._addDevice('/dev/sdc2', 29826161, 2)
        self._addDevice('/dev/sdc3', 1042177280, 3)
        self._addDevice('/dev/sdd', 11673)
        self._addDevice('/dev/sde', 1042177280)
        main.SSM_DEFAULT_BACKEND = 'btrfs'

        self.check_new_path_orig = btrfs.BtrfsPool._check_new_path
        btrfs.BtrfsPool._check_new_path = self.mock_check_new_path

    def tearDown(self):
        super(BtrfsFunctionCheck, self).tearDown()
        btrfs.BtrfsPool._check_new_path = self.check_new_path_orig

    def mock_check_new_path(self, path, name):
        pass

    def mock_run(self, cmd, *args, **kwargs):

        # Convert all parts of cmd into string
        for i, item in enumerate(cmd):
            if type(item) is not str:
                cmd[i] = str(item)

        self.run_data.append(" ".join(cmd))
        output = ""
        if cmd[:3] == ['btrfs', 'filesystem', 'show']:
            for pool, p_data in self.pool_data.iteritems():
                output += "Label: {0} uuid: some_random_uuid\n".format(pool)
                count = 0
                d_output = ""
                for dev, d_data in sorted(self.dev_data.iteritems()):
                    if 'pool_name' not in d_data or \
                       d_data['pool_name'] != pool:
                           continue
                    count += 1
                    d_output += " devid {0} size {1}.00K used {2} path {3}\n".format(
                            count, d_data['dev_size'], d_data['dev_used'], dev)
                output += " Total devices {0} FS bytes used 44.00KB\n".format(count)
                output += d_output
        elif cmd[:3] == ['btrfs', 'subvolume', 'list']:
            mpoint = cmd[-1]
            for pool, p_data in self.pool_data.iteritems():
                if 'mount' not in p_data or p_data['mount'] != mpoint:
                    continue
                count = 0
                for vol, v_data in iter(sorted(self.vol_data.iteritems())):
                    if v_data['pool_name'] != pool:
                        continue
                    count += 1
                    output += "ID {0} top level 5 path {1}\n".format(count,
                              v_data['dev_name'])
        if 'return_stdout' in kwargs and not kwargs['return_stdout']:
            output = None
        return (0, output)

    def test_btrfs_create(self):
        default_pool = btrfs.SSM_BTRFS_DEFAULT_POOL

        # Create volume using single device from non existent default pool
        self._checkCmd("ssm create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        # Specify default backend
        self._checkCmd("ssm -b btrfs create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        main.SSM_DEFAULT_BACKEND = 'lvm'
        self._checkCmd("ssm --backend btrfs create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))
        main.SSM_DEFAULT_BACKEND = 'btrfs'

        self._checkCmd("ssm -f create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        self._checkCmd("ssm -v create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        self._checkCmd("ssm -f -v create", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        self._checkCmd("ssm create", ['-s 2.6T', '/dev/sda'],
            "mkfs.btrfs -L {0} -b 2858730232217 -f /dev/sda".format(default_pool))

        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '/dev/sda'],
            "mkfs.btrfs -L btrfs_pool -m raid0 -d raid0 -b 2858730232217 -f /dev/sda".format(default_pool))
        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '/dev/sda'],
            "mkfs.btrfs -L btrfs_pool -m raid0 -d raid0 -b 2858730232217 -f /dev/sda".format(default_pool))
        self._checkCmd("ssm create", ['-r 1', '-s 512k', '/dev/sda'],
            "mkfs.btrfs -L btrfs_pool -m raid1 -d raid1 -b 524288 -f /dev/sda".format(default_pool))
        self._checkCmd("ssm create", ['-r 10', '-s 10M', '/dev/sda'],
            "mkfs.btrfs -L btrfs_pool -m raid10 -d raid10 -b 10485760 -f /dev/sda".format(default_pool))

        # Create volume using single device from non existent my_pool
        self._checkCmd("ssm create", ['--pool my_pool', '/dev/sda'],
            "mkfs.btrfs -L my_pool -f /dev/sda")

        self._checkCmd("ssm create", ['-p my_pool', '-r 0', '-s 2.6T', '/dev/sda'],
            "mkfs.btrfs -L my_pool -m raid0 -d raid0 -b 2858730232217 -f /dev/sda")

        # Create volume using multiple devices
        self._checkCmd("ssm create /dev/sda /dev/sdb", [],
            "mkfs.btrfs -L {0} -f /dev/sda /dev/sdb".format(default_pool))

        # Create volume using single device from existing pool
        self._addPool(default_pool, ['/dev/sdb', '/dev/sdd'])
        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        self._checkCmd("ssm create", ['-n myvolume'],
            "btrfs subvolume create /tmp/mount/myvolume")

        self._checkCmd("ssm create", ['-p my_pool', '-n myvolume'],
            "btrfs subvolume create /tmp/mount/myvolume")

        self._addVol('vol002', 1172832, 1, 'my_pool', ['/dev/sdc2'], '/mnt/test')
        self._checkCmd("ssm create", ['-p my_pool', '-n myvolume'],
            "btrfs subvolume create /tmp/mount/myvolume")

        # Create volume using multiple devices which one of the is in already
        # in the pool
        self._checkCmd("ssm create", ['-n myvolume', '/dev/sda /dev/sdb'],
            "btrfs subvolume create /tmp/mount/myvolume")
        self._cmdEq("btrfs device add /dev/sda /tmp/mount", -2)

        self._checkCmd("ssm create", ['-p my_pool', '-n myvolume', '/dev/sdc2 /dev/sda'],
            "btrfs subvolume create /tmp/mount/myvolume")
        self._cmdEq("btrfs device add /dev/sda /mnt/test", -2)

        self._checkCmd("ssm create", ['-n myvolume', '/dev/sda /dev/sdb /dev/sde'],
            "btrfs subvolume create /tmp/mount/myvolume")
        self._cmdEq("btrfs device add /dev/sda /dev/sde /tmp/mount", -2)

    def test_btrfs_remove(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])

        # remove volume
        self._checkCmd("ssm remove default_pool", [], "wipefs -p /dev/sda")
        self._cmdEq("wipefs -p /dev/sdb", -2)

        self._checkCmd("ssm remove my_pool", [], "wipefs -p /dev/sdc3")
        self._cmdEq("wipefs -p /dev/sdc1", -2)
        self._cmdEq("wipefs -p /dev/sdc2", -3)

        # remove subvolume
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'], '/mnt/test')
        self._checkCmd("ssm remove default_pool:/dev/default_pool/vol001", [],
            "btrfs subvolume delete /mnt/test")
        # remove device
        self._checkCmd("ssm remove /dev/sdc1", [],
            "btrfs device delete /dev/sdc1 /tmp/mount")
        # remove multiple devices
        self._checkCmd("ssm remove /dev/sdc1 /dev/sdb", [],
            "btrfs device delete /dev/sdb /mnt/test")
        self._cmdEq("btrfs device delete /dev/sdc1 /tmp/mount", -2)
        # remove combination
        self._addPool('other_pool', ['/dev/sdd', '/dev/sde'])
        self._checkCmd("ssm remove /dev/sdd /dev/sdb other_pool my_pool default_pool:/dev/default_pool/vol001", [],
            "btrfs subvolume delete /mnt/test")
        self._cmdEq("wipefs -p /dev/sdc1", -2)
        self._cmdEq("wipefs -p /dev/sdc3", -3)
        self._cmdEq("wipefs -p /dev/sdc2", -4)
        self._cmdEq("wipefs -p /dev/sde", -5)
        self._cmdEq("wipefs -p /dev/sdd", -6)
        self._cmdEq("btrfs device delete /dev/sdb /mnt/test", -7)
        self._cmdEq("btrfs device delete /dev/sdd /tmp/mount", -8)

        self._removeMount("/dev/sda")
        # remove all
        self._checkCmd("ssm remove --all", [],
            "wipefs -p /dev/sde")
        self._cmdEq("wipefs -p /dev/sdd", -2)
        self._cmdEq("wipefs -p /dev/sdc1", -3)
        self._cmdEq("wipefs -p /dev/sdc3", -4)
        self._cmdEq("wipefs -p /dev/sdc2", -5)
        self._cmdEq("wipefs -p /dev/sda", -6)
        self._cmdEq("wipefs -p /dev/sdb", -7)

        # TODO
        # remove force
        # remove verbose
        # remove verbose + force

    def test_btrfs_snapshot(self):

        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/mount1')
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'], '/mnt/mount')

        # Create snapshot
        self._checkCmd("ssm snapshot --name new_snap", ['default_pool'],
            "btrfs subvolume snapshot /mnt/mount /mnt/mount/new_snap")

        main.SSM_DEFAULT_BACKEND = 'lvm'
        self._checkCmd("ssm snapshot --name new_snap", ['default_pool'],
            "btrfs subvolume snapshot /mnt/mount /mnt/mount/new_snap")
        main.SSM_DEFAULT_BACKEND = 'btrfs'

        self._checkCmd("ssm snapshot --name new_snap", ['default_pool:/dev/default_pool/vol001'],
            "btrfs subvolume snapshot /mnt/mount /mnt/mount/new_snap")
        self._checkCmd("ssm snapshot --name new_snap", ['my_pool'],
            "btrfs subvolume snapshot /tmp/mount /tmp/mount/new_snap")

        # Create snapshot verbose
        self._checkCmd("ssm -v snapshot --name new_snap", ['default_pool'],
            "btrfs subvolume snapshot /mnt/mount /mnt/mount/new_snap")

    def test_btrfs_resize(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3'])
        self._addVol('vol001', 2982616, 1, 'my_pool', ['/dev/sdc2'],
                    '/mnt/test1')

        # Extend Volume
        self._checkCmd("ssm resize --size +4m", ['default_pool /dev/sde'],
            "btrfs filesystem resize 11723608063K /tmp/mount")

        # Specify backend
        self._checkCmd("ssm -b btrfs resize --size +4m",
            ['default_pool /dev/sde'],
            "btrfs filesystem resize 11723608063K /tmp/mount")

        main.SSM_DEFAULT_BACKEND = 'lvm'
        self._checkCmd("ssm resize --size +4m",
            ['default_pool /dev/sde'],
            "btrfs filesystem resize 11723608063K /tmp/mount")
        main.SSM_DEFAULT_BACKEND = 'btrfs'

        self._cmdEq("btrfs device add /dev/sde /tmp/mount", -2)
        self._checkCmd("ssm resize --size +1g", ['my_pool /dev/sde'],
            "btrfs filesystem resize 1073052017K /mnt/test1")
        self._cmdEq("btrfs device add /dev/sde /mnt/test1", -2)

        # Shrink volume
        self._checkCmd("ssm resize", ['-s-100G', 'default_pool'],
            "btrfs filesystem resize 11618746367K /tmp/mount")
        self._checkCmd("ssm resize -s-500G", ['my_pool /dev/sde'],
            "btrfs filesystem resize 547715441K /mnt/test1")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sde /mnt/test1")

        # Set volume size
        self._checkCmd("ssm resize", ['-s 10M', 'default_pool'],
            "btrfs filesystem resize 10240K /tmp/mount")
        self._checkCmd("ssm resize", ['-s 10M', 'my_pool'],
            "btrfs filesystem resize 10240K /mnt/test1")

        # Set volume and add devices
        self._checkCmd("ssm resize -s 12T default_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 12884901888K /tmp/mount")
        self.assertEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")
        self._checkCmd("ssm resize -s 1T my_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 1073741824K /mnt/test1")
        self.assertEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /mnt/test1")
        self._checkCmd("ssm resize -s 1T my_pool /dev/sde /dev/sdc2",
            [], "btrfs filesystem resize 1073741824K /mnt/test1")
        self.assertEqual(self.run_data[-2],
            "btrfs device add /dev/sde /mnt/test1")

        # Set volume in without the need adding more devices
        self._checkCmd("ssm resize -s 10G default_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 10485760K /tmp/mount")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")
        self._checkCmd("ssm resize -s 10G my_pool /dev/sdd /dev/sde",
            [], "btrfs filesystem resize 10485760K /mnt/test1")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /mnt/test1")

        # Extend volume and add devices
        self._checkCmd("ssm resize -s +500G default_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 12247891967K /tmp/mount")
        self.assertEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")

        # Extend volume in without the need adding more devices
        self._checkCmd("ssm resize -s 1k default_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 1K /tmp/mount")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")

        # Shrink volume with devices provided
        self._checkCmd("ssm resize -s-10G default_pool /dev/sdc1 /dev/sde",
            [], "btrfs filesystem resize 11713118207K /tmp/mount")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")
        self.assertNotEqual(self.run_data[-2],
            "btrfs device add /dev/sdc1 /dev/sde /tmp/mount")

    def test_btrfs_add(self):
        default_pool = btrfs.SSM_BTRFS_DEFAULT_POOL

        # Adding to non existent pool
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        # Specify backend
        self._checkCmd("ssm --backend btrfs add", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))

        main.SSM_DEFAULT_BACKEND = 'lvm'
        self._checkCmd("ssm -b btrfs add", ['/dev/sda'],
            "mkfs.btrfs -L {0} -f /dev/sda".format(default_pool))
        main.SSM_DEFAULT_BACKEND = 'btrfs'

        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "mkfs.btrfs -L {0} -f /dev/sda /dev/sdc1".format(default_pool))
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "mkfs.btrfs -L my_pool -f /dev/sda")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "mkfs.btrfs -L my_pool -f /dev/sda")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "mkfs.btrfs -L my_pool -f /dev/sda /dev/sdc1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "mkfs.btrfs -L my_pool -f /dev/sda /dev/sdc1")

        # Adding to existing default pool
        self._addPool(default_pool, ['/dev/sdb', '/dev/sdd'])
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "btrfs device add /dev/sda /tmp/mount")
        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "btrfs device add /dev/sda /dev/sdc1 /tmp/mount")

        # Adding to existing defined pool
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3'])
        self._addVol('vol001', 2982616, 1, 'my_pool', ['/dev/sdc2'],
                    '/mnt/test1')
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "btrfs device add /dev/sda /mnt/test1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "btrfs device add /dev/sda /mnt/test1")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "btrfs device add /dev/sda /dev/sdc1 /mnt/test1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "btrfs device add /dev/sda /dev/sdc1 /mnt/test1")
        # Add verbose
        self._checkCmd("ssm -v add", ['--pool {0}'.format(default_pool),
            '/dev/sda /dev/sdc1'],
            "btrfs device add /dev/sda /dev/sdc1 /tmp/mount")

        # Add two devices into existing pool (one of the devices already is in
        # the pool
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sdc2 /dev/sda'],
            "btrfs device add /dev/sda /mnt/test1")
        self._checkCmd("ssm add", ['/dev/sda /dev/sdb'],
            "btrfs device add /dev/sda /tmp/mount")

    def test_btrfs_mount(self):
        self._addDir("/mnt/test")
        self._addDir("/mnt/test1")
        self._addDir("/mnt/test2")
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/mount1')
        self._addVol('vol003/vol006', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'], '/mnt/mount')

        # Mount subvolume
        main.main("ssm mount default_pool:/dev/default_pool/vol002 /mnt/test")
        self._cmdEq("mount -o subvolid=2 /dev/sda /mnt/test")
        main.main("ssm mount -o discard default_pool:/dev/default_pool/vol002 /mnt/test")
        self._cmdEq("mount -o discard,subvolid=2 /dev/sda /mnt/test")
        main.main("ssm mount --options rw,discard,neco=44 default_pool:/dev/default_pool/vol004 /mnt/test1")
        self._cmdEq("mount -o rw,discard,neco=44,subvolid=4 /dev/sda /mnt/test1")

        main.main("ssm mount -o discard default_pool:/dev/default_pool/vol003/vol006 /mnt/test2")
        self._cmdEq("mount -o discard,subvolid=3 /dev/sda /mnt/test2")

        # Mount the whole file system
        main.main("ssm mount default_pool /mnt/test2")
        self._cmdEq("mount -o subvolid=0 /dev/sda /mnt/test2")
        main.main("ssm mount --options rw,discard,neco=44 default_pool /mnt/test1")
        self._cmdEq("mount -o rw,discard,neco=44,subvolid=0 /dev/sda /mnt/test1")
        main.main("ssm mount /dev/sdb /mnt/test1")
        self._cmdEq("mount /dev/sdb /mnt/test1")
        main.main("ssm mount -o rw,discard,neco=44 /dev/sdb /mnt/test1")
        self._cmdEq("mount -o rw,discard,neco=44 /dev/sdb /mnt/test1")

        # Non existing volume
        main.main("ssm mount nonexisting /mnt/test1")
        self._cmdEq("mount nonexisting /mnt/test1")
        main.main("ssm mount -o discard,rw nonexisting /mnt/test1")
        self._cmdEq("mount -o discard,rw nonexisting /mnt/test1")



import os
import sys
from distutils.core import setup
from ssmlib import main

VERSION=main.VERSION
DOC_BUILD='doc/_build/'
NAME="system-storage-manager"

if sys.version < '2.6':
    print "Python version 2.6 or higher is required " + \
         "for System Storage Manager to run correctly!"
    sys.exit(1)

setup(
    name=NAME,
    version=VERSION,
    author='Lukas Czerner',
    author_email='lczerner@redhat.com',
    maintainer='Lukas Czerner',
    maintainer_email='lczerner@redhat.com',
    url='http://storagemanager.sf.net',
    packages=['ssmlib', 'ssmlib.backends'],
    scripts=['bin/ssm'],
    description='System Storage Manager - A single tool to manage your storage',
    license='GNU General Public License version 2 or any later version',
    long_description=open('README').read(),
    requires=['argparse', 'itertools', 'tempfile', 'threading', 'subprocess',
              'datetime', 're', 'os', 'sys', 'stat'],
    platforms=['Linux'],
    data_files=[('/usr/share/man/man8', ['doc/_build/man/ssm.8']),
                ('/usr/share/doc/{0}-{1}'.format(NAME, VERSION),
                    ['README', 'CHANGES', 'COPYING', 'AUTHORS', 'INSTALL'])]
)

